"""
Create a map from one column of a stacked Jacobian to a full (dense or
sparse) Jacobian matrix
"""


#[
from __future__ import annotations

from typing import (Self, Any, )
from collections.abc import (Iterable, )
import itertools as it_
import dataclasses as dc_
#]


@dc_.dataclass
class ArrayMap:
    """
    """
    lhs: tuple[list[int], list[int]] | None = None
    rhs: tuple[list[int], list[int]] | None = None
    #[
    def __init__(self, /, ) -> None:
        self.lhs = ([], [])
        self.rhs = ([], [])

    def __len__(self, /, ) -> int:
        return len(self.lhs[0])

    def append(
        self,
        lhs: tuple[int, int], 
        rhs: tuple[int, int],
        /
    ) -> None:
        """
        """
        self.lhs[0].append(lhs[0])
        self.lhs[1].append(lhs[1])
        self.rhs[0].append(rhs[0])
        self.rhs[1].append(rhs[1])

    def merge_with(
        self,
        other: Self,
    ) -> None:
        """
        """
        self.lhs = (self.lhs[0]+other.lhs[0], self.lhs[1]+other.lhs[1])
        self.rhs = (self.rhs[0]+other.rhs[0], self.rhs[1]+other.rhs[1])

    def remove_nones(self) -> None:
        """
        Remove any map entry that has a None for the row index on the LHS
        """
        if not self.lhs[0]:
            return
        zipped_pruned = [
            i for i in zip(self.lhs[0], self.lhs[1], self.rhs[0], self.rhs[1])
            if i[0] is not None
        ]
        unzipped_pruned = list(zip(*zipped_pruned))
        self.lhs = (list(unzipped_pruned[0]), list(unzipped_pruned[1]))
        self.rhs = (list(unzipped_pruned[2]), list(unzipped_pruned[3]))

    @classmethod
    def for_equations(
        cls,
        eids: list[int],
        eid_to_wrt_tokens: dict[int, Any],
        tokens_in_columns_on_rhs: list[Any],
        eid_to_rhs_offset: dict[int, int],
        /,
        **kwargs,
    ) -> Self:
        """
        """
        return merge_array_maps(
            cls._for_single_equation(
                eid_to_wrt_tokens[eid],
                tokens_in_columns_on_rhs,
                eid_to_rhs_offset[eid],
                lhs_row,
                **kwargs,
            )
            for lhs_row, eid in enumerate(eids)
        )

    @classmethod
    def _for_single_equation(
        cls,
        tokens_in_equation_on_rhs: in_.Tokens,
        tokens_in_columns_on_lhs: in_.Tokens,
        rhs_offset: int,
        lhs_row: int,
        /,
        rhs_column: int,
        lhs_column_offset: int,
    ) -> Self:
        """
        """
        index_func = lambda t: lhs_column_offset + tokens_in_columns_on_lhs.index(t)
        raw_map = (
            (lhs_row, index_func(t), rhs_row, rhs_column) 
            for rhs_row, t in enumerate(tokens_in_equation_on_rhs, start=rhs_offset)
            if t in tokens_in_columns_on_lhs
        )
        # Collect all lhr_rows, all lhs_columns, all rhs_rows, all rhs_columns
        raw_map = zip(*raw_map)
        self = cls()
        try:
            self.lhs = (list(next(raw_map)), list(next(raw_map)))
            self.rhs = (list(next(raw_map)), list(next(raw_map)))
        except:
            self.lhs = ([], [])
            self.rhs = ([], [])
        # Equivalent to:
        # self = cls()
        # for rhs_row, t in enumerate(tokens_in_equation_on_rhs, start=rhs_offset):
        #     if t in tokens_in_columns_on_lhs:
        #         lhs_column = tokens_in_columns_on_lhs.index_func(t)
        #         self.append((lhs_row, lhs_column), (rhs_row, 0))
        return self

    @classmethod
    def constant_vector(
        cls,
        eids: Iterable[int],
    ) -> Self:
        """
        """
        num_equations = len(eids)
        self = cls()
        self.lhs = (list(range(num_equations)), [0]*num_equations)
        self.rhs = (list(eids), [0]*num_equations)
        return self
    #]


def merge_array_maps(maps: Iterable[ArrayMap]) -> ArrayMap:
    """
    """
    #[
    stacked_map = ArrayMap()
    for m in maps:
        stacked_map.merge_with(m)
    return stacked_map
    #]


def create_eid_to_rhs_offset(
    eids: tuple[int],
    eid_to_wrt_something: dict[int, Any],
    /,
) -> dict[int, int]:
    """
    Cumulative sum of number of tokens in individual equations, starting with 0
    """
    rhs_offsets = list(it_.accumulate(
        len(eid_to_wrt_something[i]) 
        for i in eids
    ))
    # Offset starts from 0 for the first equations
    rhs_offsets.pop()
    rhs_offsets.insert(0, 0)
    return dict(zip(eids, rhs_offsets))





"""
Algebraic source
"""

#[
from __future__ import annotations

import re as _re
from typing import (TypeAlias, Literal, )
from collections.abc import (Iterable, )

from .parsers import preparser as _preparser
from .parsers import algebraic as _algebraic
from .parsers import common as _common
from . import equations as _equations
from . import quantities as _quantities
#]


QuantityInput: TypeAlias = tuple[str, str]
EquationInput: TypeAlias = tuple[str, tuple[str, str]]


LOGLY_VARIABLE = (
    _quantities.QuantityKind.TRANSITION_VARIABLE
    | _quantities.QuantityKind.MEASUREMENT_VARIABLE
    | _quantities.QuantityKind.EXOGENOUS_VARIABLE
)


class SourceMixin:
    """
    """
    #[
    @classmethod
    def from_string(
        cls,
        source_string: str,
        /,
        custom_functions: dict | None = None,
        save_preparsed: str = "",
        **kwargs,
    ) -> Self:
        """
        Create a new object from algebraic source string
        """
        source, info = AlgebraicSource.from_string(
            source_string, context=custom_functions, save_preparsed=save_preparsed,
        )
        return cls.from_source(source, context=custom_functions, **kwargs, )

    @classmethod
    def from_file(
        cls,
        source_files: str | Iterable[str],
        /,
        **kwargs,
    ) -> Self:
        """
        Create a new object from algebraic source files
        """
        source_string = _common.combine_source_files(source_files, )
        return cls.from_string(source_string, **kwargs, )
    #]


class AlgebraicSource:
    """
    """
    #[
    __slots__ = (
        "quantities", "dynamic_equations", "steady_equations",
        "log_variables", "all_but", "context", "shock_qid_to_std_qid",
    )
    def __init__(self, /) -> None:
        self.quantities = []
        self.dynamic_equations = []
        self.steady_equations = []
        self.log_variables = []
        self.all_but = []
        self.context = None
        self.shock_qid_to_std_qid = None

    @property
    def num_quantities(self, /) -> int:
        return len(self.quantities)

    @property
    def all_names(self, /) -> Iterable[str]:
        return [ qty.human for qty in self.quantities ]

    def add_parameters(self, names: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(names, _quantities.QuantityKind.PARAMETER)

    def add_exogenous_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.EXOGENOUS_VARIABLE)

    def add_transition_variables(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.TRANSITION_VARIABLE)

    def add_transition_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.TRANSITION_SHOCK)

    def add_transition_equations(self, equation_inputs: Iterable[EquationInput] | None) -> None:
        self._add_equations(equation_inputs, _equations.EquationKind.TRANSITION_EQUATION)

    def add_measurement_variables(self, quantity_inputs: Iterable[EquationInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.MEASUREMENT_VARIABLE)

    def add_measurement_shocks(self, quantity_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_quantities(quantity_inputs, _quantities.QuantityKind.MEASUREMENT_SHOCK)

    def add_measurement_equations(self, equation_inputs: Iterable[QuantityInput] | None) -> None:
        self._add_equations(equation_inputs, _equations.EquationKind.MEASUREMENT_EQUATION)

    def add_all_but(self, all_but_input: str | None) -> None:
        if not all_but_input:
            return
        self.all_but.append(all_but_input)

    def add_log_variables(self, log_variable_inputs: Iterable[str] | None, /, ) -> None:
        if not log_variable_inputs:
            return
        self.log_variables += log_variable_inputs

    def _add_quantities(self, quantity_inputs: Iterable[QuantityInput] | None, kind: _quantities.QuantityKind) -> None:
        if not quantity_inputs:
            return
        offset = self.quantities[-1].entry + 1 if self.quantities else 0
        self.quantities = self.quantities + [
            _quantities.Quantity(id=None, human=q[1].strip(), kind=kind, logly=None, description=q[0].strip(), entry=i)
            for i, q in enumerate(quantity_inputs, start=offset)
        ]

    def _add_equations(self, equation_inputs: Iterable[EquationInput] | None, kind: _equations.EquationKind, /) -> None:
        if not equation_inputs:
            return
        offset = self.dynamic_equations[-1].entry + 1 if self.dynamic_equations else 0
        self.dynamic_equations = self.dynamic_equations + [
            _equations.Equation(id=None, human=_handle_white_spaces(ein[1][0]), kind=kind, description=ein[0].strip(), entry=i)
            for i, ein in enumerate(equation_inputs, start=offset)
        ]
        self.steady_equations = self.steady_equations + [
            _equations.Equation(id=None, human=_handle_white_spaces(ein[1][1] if ein[1][1] else ein[1][0]), kind=kind, description=ein[0].strip(), entry=i)
            for i, ein in enumerate(equation_inputs, start=offset)
        ]

    def populate_logly(self, /) -> None:
        """
        """
        #
        # Logly status for nonlisted variables
        default_logly = self._resolve_default_logly()
        #
        # Logly status for listed variables
        listed_logly = not default_logly
        #
        for qty in self.quantities:
            if qty.kind not in LOGLY_VARIABLE:
                continue
            qty.logly = listed_logly if qty.human in self.log_variables else default_logly

    def _is_logly_consistent(self, /) -> bool:
        return all(a=="all-but" for a in self.all_but) or all(a=="" for a in self.all_but) if self.all_but else True

    def _resolve_default_logly(self, /) -> bool:
        if not self._is_logly_consistent():
            raise Exception("Inconsistent use of !all-but in !log-variables")
        return self.all_but.pop()=="all-but" if self.all_but else False

    @classmethod
    def from_lists(
        cls,
        /,
        transition_variables: Iterable[QuantityInput],
        transition_equations: Iterable[EquationInput], 
        transition_shocks: Iterable[QuantityInput] | None = None,
        measurement_variables: Iterable[QuantityInput] | None = None,
        measurement_equations: Iterable[EquationInput] | None = None,
        measurement_shocks: Iterable[QuantityInput] | None = None,
        parameters: Iterable[QuantityInput] | None = None,
        exogenous_variables: Iterable[QuantityInput] | None = None,
        log_variables: Iterable[str] | None = None,
    ) -> Self:
        """
        """
        self = AlgebraicSource()
        self.add_transition_variables(transition_variables)
        self.add_transition_equations(transition_equations)
        self.add_transition_shocks(transition_shocks)
        self.add_measurement_variables(measurement_variables)
        self.add_measurement_equations(measurement_equations)
        self.add_measurement_shocks(measurement_shocks)
        self.add_parameters(parameters)
        self.add_exogenous_variables(exogenous_variables)
        self.add_log_variables(log_variables)
        self.populate_logly()
        return self

    @classmethod
    def from_string(
        cls,
        source_string: str,
        /,
        context: dict | None = None,
        save_preparsed: str = "",
    ) -> tuple[Self, dict]:
        """
        """
        preparsed_string, preparser_info = _preparser.from_string(
            source_string, context, save_preparsed=save_preparsed, 
        )
        parsed_content = _algebraic.from_string(preparsed_string)
        #
        self = AlgebraicSource()
        self.add_transition_variables(parsed_content.get("transition-variables"))
        self.add_transition_shocks(parsed_content.get("transition-shocks"))
        self.add_transition_equations(parsed_content.get("transition-equations"))
        self.add_measurement_variables(parsed_content.get("measurement-variables"))
        self.add_measurement_shocks(parsed_content.get("measurement-shocks"))
        self.add_measurement_equations(parsed_content.get("measurement-equations"))
        self.add_parameters(parsed_content.get("parameters"))
        self.add_exogenous_variables(parsed_content.get("exogenous-variables"))
        self.add_log_variables(parsed_content.get("log-variables"))
        for i in parsed_content.get("all-but", []):
            self.add_all_but(i)
        self.context = preparser_info["context"]
        self.populate_logly()
        #
        return self, preparser_info
    #]


def _handle_white_spaces(x: str, /) -> str:
    return _re.sub(r"[\s\n\r]+", "", x)




"""
"""


#[
from __future__ import annotations

from typing import (Protocol, )
from numbers import (Number, )
import re as _re
import numpy as _np

from .incidences import main as _incidence
#]


class TransformProtocol(Protocol, ):
    _LHS_PATTERN: _re.Pattern
    def __str__(self) -> str: ...


class Transform:
    """
    """
    #[
    def __init__(
        self,
        /,
        when_data: bool = False,
    ) -> None:
        self.when_data = when_data

    def __str__(self, /, ) -> str:
        when_data_symbol = "!" if not self.when_data else "?"
        return when_data_symbol + self._SYMBOL

    def __repr__(self, /, ) -> self:
        return self.__str__()

    def resolve_databank_name(
        self,
        base_name: str, 
        /,
    ) -> str:
        return self._DATABANK_NAME.replace("$", base_name, )
    #]


class TransformNone(Transform, ):
    """
    """
    #[
    _LHS_PATTERN = _re.compile(r"(\w+)")
    _DATABANK_NAME = "$"
    _SYMBOL = "••"

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        return rhs_xtring

    def eval_exogenized(
        self,
        exogenized_value: Number,
        lagged: Number,
        /,
    ) -> Number:
        return exogenized_value
    #]


class TransformLog(Transform, ):
    """
    """
    #[
    _LHS_PATTERN = _re.compile(r"log\((\w+)\)")
    _DATABANK_NAME = "log_$"
    _SYMBOL = "log"

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        return f"exp({rhs_xtring})"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        lagged: Number,
        /,
    ) -> Number:
        return _np.exp(exogenized_value)
    #]


class TransformDiff(Transform, ):
    """
    """
    #[
    _LHS_PATTERN = _re.compile(r"\(\((\w+)\)-\(\1\[-1\]\)\)")
    _DATABANK_NAME = "diff_$"
    _SYMBOL = "diff"

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        lhs_token_lag_printed = lhs_token.shifted(-1, ).print_xtring()
        return f"{lhs_token_lag_printed}+({rhs_xtring})"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        lagged: Number,
        /,
    ) -> Number:
        return lagged + exogenized_value
    #]


class TransformDiffLog(Transform, ):
    """
    """
    #[
    _LHS_PATTERN = _re.compile(r"\(log\((\w+)\)-log\(\1\[-1\]\)\)")
    _DATABANK_NAME = "difflog_$"
    _SYMBOL = "difflog"

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        lhs_token_lag_printed = lhs_token.shifted(-1, ).print_xtring()
        return f"{lhs_token_lag_printed}*exp({rhs_xtring})"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        lagged: Number,
        /,
    ) -> Number:
        return lagged * _np.exp(exogenized_value)
    #]


class TransformRoc(Transform, ):
    """
    """
    #[
    _LHS_PATTERN = _re.compile(r"\(\((\w+)\)/\(\1\[-1\]\)\)")
    _DATABANK_NAME = "roc_$"
    _SYMBOL = "roc"

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        lhs_token_lag_printed = lhs_token.shifted(-1, ).print_xtring()
        return f"{lhs_token_lag_printed}*({rhs_xtring})"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        lagged: Number,
        /,
    ) -> Number:
        return lagged * exogenized_value
    #]


class TransformPct(Transform, ):
    """
    """
    #[
    _LHS_PATTERN = _re.compile(r"\(100*\((\w+)\)/\(\1\[-1\]\)-100\)")
    _DATABANK_NAME = "pct_$"
    _SYMBOL = "pct"

    def create_eval_level_str(
        self,
        lhs_token: _incidence.Token,
        rhs_xtring: str,
        /,
    ) -> str:
        lhs_token_lag_printed = lhs_token.shifted(-1, ).print_xtring()
        return f"{lhs_token_lag_printed}*(1+({rhs_xtring})/100)"

    def eval_exogenized(
        self,
        exogenized_value: Number,
        lagged: Number,
        /,
    ) -> Number:
        return lagged * (1+exogenized_value/100)
    #]


_ALL_TRANSFORMS = (
    TransformNone,
    TransformLog,
    TransformDiff,
    TransformDiffLog,
    TransformRoc,
    TransformPct,
)


RESOLVE_TRANSFORM = {
    None: TransformNone,
    "level": TransformNone,
    "none": TransformNone,
    "log": TransformLog,
    "diff": TransformDiff,
    "difflog": TransformDiffLog,
    "roc": TransformRoc,
    "pct": TransformPct,
}


def recognize_transform_in_equation(
    lhs: str,
    /,
) -> tuple[TransformProtocol | None, str | None]:
    """
    """
    #[
    for t in _ALL_TRANSFORMS:
        m = t._LHS_PATTERN.fullmatch(lhs)
        if m is not None:
            return t(), m.group(1)
    return None, None
    #]





import sys
sys.path.append("../../../src")

import random as _ra

from irispie import *

#x = Series.from_func(qq(2020,1)>>qq(2025,4), _ra.gauss, num_columns=2)
x = 10+Series.from_func(qq(2020,1)>>qq(2025,4), _ra.gauss, )
#x = exp(x)

start_date = qq(2020,1)
end_date = qq(2025,4) + 8
change = Series.from_start_date_and_values(end_date, 0)
xt, xg = x.hpf(log=False, range=start_date>>end_date, )#change=change, )




"""
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, )
from collections.abc import (Iterable, Callable, )
from numbers import (Number, )
import copy as _co
import numpy as _np
import itertools as _it
import functools as _ft

from .. import equations as _equations
from .. import quantities as _quantities
from .. import sources as _sources
from .. import dates as _dates
from ..parsers import common as _pc
from ..databanks import main as _databanks
from ..fords import solutions as _solutions
from ..fords import steadiers as _fs
from ..fords import descriptors as _descriptors
from ..fords import systems as _systems
from .. import wrongdoings as _wrongdoings

from . import variants as _variants
from . import invariants as _invariants
from . import _covariances as _covariances
from . import _flags as _flags
from . import _simulate as _simulate
from . import _steady as _steady
from . import _get as _get
from . import _processor as _processor
#]


__all__ = [
    "Model"
]


class Model(
    _sources.SourceMixin,
    _simulate.SimulateMixin,
    _steady.SteadyMixin,
    _covariances.CoverianceMixin,
    _get.GetMixin,
    _processor.ProcessorMixin,
):
    """

    """
    #[
    __slots__ = (
        "_invariant", "_variants",
    )

    def __init__(
        self,
        /,
    ) -> None:
        """
        """
        self._invariant = None
        self._variants = None

    def assign(
        self,
        /,
        **kwargs,
    ) -> None:
        """
        """
        name_to_qid = self.create_name_to_qid()
        qid_to_name = self.create_qid_to_name()
        #
        qid_to_value = _rekey_dict(kwargs, name_to_qid, )
        for vid, variant in enumerate(self._variants, ):
            qid_to_value_variant = _extract_dict_variant(qid_to_value, vid, qid_to_name, )
            variant.update_values_from_dict(qid_to_value_variant, )
            self._enforce_auto_values(variant, )

    def copy(self) -> Self:
        """
        Create a deep copy of the Model object
        """
        return _co.deepcopy(self)

    def __getitem__(
        self,
        request,
        /,
    ):
        """
        """
        if isinstance(request, str):
            return self._getitem_value(request, )
        else:
            return self._getitem_variant(variant, )

    def _getitem_value(
        self,
        name: str,
        /,
    ) -> Any:
        """
        """
        names = (name, )
        quantities = self._invariant.quantities
        qids, invalid_names = _quantities.lookup_qids_by_name(quantities, names, )
        if invalid_names:
            raise _wrongdoings.IrisPieError(f"Invalid model name \"{invalid_names[0]}\"", )
        else:
            singleton_dict = self._get_values("levels", qids, )
            return singleton_dict[name]

    def _getitem_variant(
        self,
        variants,
        /,
    ) -> Self:
        new = self.from_self()
        index_variants = resolve_variant(self, variants, )
        new._variants = [ self._variants[i] for i in index_variants ]
        return new

    def alter_num_variants(
        self,
        new_num: int,
        /,
    ) -> Self:
        """
        Alter (expand, shrink) the number of alternative parameter variants in this model object
        """
        if new_num < self.num_variants:
            self._shrink_num_variants(new_num, )
        elif new_num > self.num_variants:
            self._expand_num_variants(new_num, )
        return self

    @property
    def is_singleton(self, /, ) -> bool:
        """
        True for Models with only one variant
        """
        return self.num_variants == 1

    def change_logly(
        self,
        new_logly: bool,
        some_names: Iterable[str] | None = None,
        /
    ) -> None:
        """
        Change the log-status of some Model quantities
        """
        some_names = set(some_names) if some_names else None
        qids = [
            qty.id
            for qty in self._invariant.quantities
            if qty.logly is not None and (some_names is None or qty.human in some_names)
        ]
        self._invariant.quantities = _quantities.change_logly(self._invariant.quantities, new_logly, qids)

    @property
    def num_variants(self, /, ) -> int:
        """
        Number of alternative variants within this Model
        """
        return len(self._variants)

    @property
    def is_linear(self, /, ) -> bool:
        """
        True for Models declared as linear
        """
        return self._invariant._flags.is_linear

    @property
    def is_flat(self, /, ) -> bool:
        """
        True for Models declared as flat
        """
        return self._invariant._flags.is_flat

    def create_name_to_qid(self, /, ) -> dict[str, int]:
        return _quantities.create_name_to_qid(self._invariant.quantities)

    def create_qid_to_name(self, /, ) -> dict[int, str]:
        return _quantities.create_qid_to_name(self._invariant.quantities)

    def create_qid_to_kind(self, /, ) -> dict[int, str]:
        return _quantities.create_qid_to_kind(self._invariant.quantities)

    def create_qid_to_description(self, /, ) -> dict[int, str]:
        """
        Create a dictionary mapping from quantity id to quantity descriptor
        """
        return _quantities.create_qid_to_description(self._invariant.quantities)

    def create_qid_to_logly(self, /, ) -> dict[int, bool]:
        """
        Create a dictionary mapping from quantity id to quantity log-status
        """
        return _quantities.create_qid_to_logly(self._invariant.quantities)

    def create_steady_array(
        self,
        /,
        variant: _variants.Variant|None = None,
        **kwargs,
    ) -> _np.ndarray:
        qid_to_logly = self.create_qid_to_logly()
        if variant is None:
            variant = self._variants[0]
        return variant.create_steady_array(qid_to_logly, **kwargs, )

    def create_zero_array(
        self,
        /,
        variant: _variants.Variant|None = None,
        **kwargs,
    ) -> _np.ndarray:
        """
        """
        qid_to_logly = self.create_qid_to_logly()
        if variant is None:
            variant = self._variants[0]
        return variant.create_zero_array(qid_to_logly, **kwargs, )

    def create_some_array(
        self,
        /,
        deviation: bool,
        **kwargs,
    ) -> _np.ndarray:
        return {
            True: self.create_zero_array, False: self.create_steady_array,
        }[deviation](**kwargs)

    def _enforce_auto_values(self, variant, /, ) -> None:
        """
        """
        #
        # Reset levels of shocks to zero, remove changes
        #
        assign_shocks = {
            qid: (0, _np.nan)
            for qid in _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.SHOCK)
        }
        variant.update_values_from_dict(assign_shocks)
        #
        # Remove changes from quantities that are not logly variables
        #
        assign_non_logly = {
            qid: (..., _np.nan)
            for qid in _quantities.generate_qids_by_kind(self._invariant.quantities, ~_sources.LOGLY_VARIABLE)
        }
        variant.update_values_from_dict(assign_non_logly)

    def _shrink_num_variants(self, new_num: int, /, ) -> None:
        """
        """
        if new_num<1:
            raise Exception('Number of variants must be one or more')
        self._variants = self._variants[0:new_num]

    def _expand_num_variants(self, new_num: int, /, ) -> None:
        """
        """
        for i in range(self.num_variants, new_num):
            self._variants.append(_co.deepcopy(self._variants[-1]))

    def systemize(
        self,
        /,
        **kwargs,
    ) -> Iterable[_systems.System]:
        """
        Create unsolved first-order system for each variant
        """
        model_flags = self._invariant._flags.update_from_kwargs(**kwargs, )
        return tuple(
            self._systemize(variant, self._invariant.dynamic_descriptor, model_flags, )
            for variant in self._variants
        )

    def _systemize(
        self,
        variant: _variants.Variant,
        descriptor: _descriptors.Descriptor,
        model_flags: flags.Flags,
        /,
    ) -> _systems.System:
        """
        Create unsolved first-order system for one variant
        """
        ac = descriptor.aldi_context
        num_columns = ac.shape_data[1]
        qid_to_logly = self.create_qid_to_logly()
        #
        if model_flags.is_linear:
            data_array = variant.create_zero_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=ac.min_shift, )
            data_array_lagged = None
            steady_array = variant.create_steady_array(qid_to_logly, num_columns=1, ).reshape(-1)
        else:
            data_array = variant.create_steady_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=ac.min_shift, )
            data_array_lagged = variant.create_steady_array(qid_to_logly, num_columns=num_columns, shift_in_first_column=ac.min_shift-1, )
            steady_array = data_array[:, -ac.min_shift]
        #
        return _systems.System(descriptor, data_array, steady_array, model_flags, data_array_lagged, )

    def solve(
        self,
        /,
        **kwargs,
    ) -> None:
        """
        Calculate first-order solution for each Variant within this Model
        """
        model_flags = self._invariant._flags.update_from_kwargs(**kwargs, )
        for variant in self._variants:
            self._solve(variant, model_flags, )

    def _solve(
        self,
        variant: _variants.Variant,
        model_flags: flags.Flags,
        /,
    ) -> None:
        """
        Calculate first-order solution for one Variant of this Model
        """
        system = self._systemize(variant, self._invariant.dynamic_descriptor, model_flags, )
        variant.solution = _solutions.Solution(self._invariant.dynamic_descriptor, system, )

    def _choose_plain_equator(
        self,
        equation_switch: Literal["dynamic", "steady", ],
        /,
    ) -> Callable | None:
        """
        """
        match equation_switch:
            case "dynamic":
                return self._invariant._plain_equator_for_dynamic_equations
            case "steady":
                return self._invariant._plain_equator_for_steady_equations

    def _assign_default_stds(
        self,
        default_std: Number | None = None,
        /,
    ) -> None:
        """
        """
        default_std = default_std if default_std is not None else _DEFAULT_STD_RESOLUTION(self.get_flags())
        dict_to_assign = {
            k: default_std
            for k in _quantities.generate_quantity_names_by_kind(self._invariant.quantities, _quantities.QuantityKind.STD, )
        }
        self.assign(**dict_to_assign, )

    @classmethod
    def from_source(
        cls,
        source: _sources.AlgebraicSource,
        /,
        default_std: int | None = None,
        context: dict | None = None,
        **kwargs,
    ) -> Self:
        """
        """
        self = cls()
        self._invariant = _invariants.Invariant(
            source,
            context=context,
            **kwargs,
        )
        self._variants = [ _variants.Variant(self._invariant.quantities, ) ]
        for v in self._variants:
            self._enforce_auto_values(v, )
        self._assign_default_stds(default_std, )
        return self

    def from_self(self, ) -> Self:
        """
        Create a new Model object with pointers to invariant and variants of this Model object
        """
        new = type(self)()
        new._invariant = self._invariant
        new._variants = self._variants
        return new

    #
    # Implement SimulatableProtocol
    #

    def get_min_max_shifts(self) -> tuple[int, int]:
        """
        """
        return self._invariant._min_shift, self._invariant._max_shift

    def get_databank_names(self, plan, /, ) -> tuple[str, ...]:
        qid_to_name = self.create_qid_to_name()
        return tuple(qid_to_name[qid] for qid in range(len(qid_to_name)))
    #]


_DEFAULT_STD_LINEAR = 1
_DEFAULT_STD_NONLINEAR = 0.01
_DEFAULT_STD_RESOLUTION = lambda flags: {
    True: _DEFAULT_STD_LINEAR,
    False: _DEFAULT_STD_NONLINEAR,
}[flags.is_linear]


def _rekey_dict(dict_to_rekey: dict, old_key_to_new_key: dict, /, garbage_key=None, ) -> dict:
    #[
    new_dict = {
        old_key_to_new_key.get(key, garbage_key): value
        for key, value in dict_to_rekey.items()
    }
    if garbage_key in new_dict:
        del new_dict[garbage_key]
    return new_dict
    #]


def resolve_variant(self, variants, /, ) -> Iterable[int]:
    #[
    if isinstance(variants, Number):
        return [variants, ]
    elif variants is Ellipsis:
        return range(self.num_variants)
    elif isinstance(variants, slice):
        return range(*variants.indices(self.num_variants))
    else:
        return [v for v in variants]
    #]


def _extract_dict_variant(
    qid_to_value: dict[int, Any],
    vid: int,
    qid_to_name: dict[int, str],
    /,
) -> dict[int, tuple[Number|Ellipsis, Number|Ellipsis]]:
    """
    """
    def _extract_value_variant(
        value: list | tuple | Number,
        name: str,
        /,
    ) -> tuple[Number|Ellipsis, Number|Ellipsis]:
        """
        """
        if isinstance(value, Iterable, ) and not isinstance(value, str, ) and not isinstance(value, tuple, ):
            value = list(value, )[vid]
        if isinstance(value, Number, ) or value is Ellipsis:
            value = (value, Ellipsis, )
        if not isinstance(value, tuple, ) or len(value) != 2:
            raise _wrongdoings.IrisPieError(
                f"Invalid type of value assigned to this Model quantity: \"{name}\"",
            )
        return value
        #
    return {
        qid: _extract_value_variant(value, qid_to_name[qid], )
        for qid, value in qid_to_value.items()
    }



"""
Iris Pie
"""

from .dates import *
from .dates import __all__ as dates_all

from .series import *
from .series import __all__ as series_all

from .databanks.main import *
from .databanks.main import __all__ as databanks_all

from .dataslabs import *
from .dataslabs import __all__ as dataslabs_all

from .models import *
from .models import __all__ as models_all

from .quantities import *
from .quantities import __all__ as quantities_all

from .equations import *
from .equations import __all__ as equations_all

from .sequentials import *
from .sequentials import __all__ as sequentials_all

from .explanatories import *
from .explanatories import __all__ as explanatories_all

from .plans.main import *
from .plans.main import __all__ as plans_all

from .namings import *
from .namings import __all__ as namings_all


__all__ = (
    *dates_all,
    *series_all,
    *databanks_all,
    *models_all,
    *quantities_all,
    *equations_all,
    *sequentials_all,
    *explanatories_all,
    *plans_all,
    *namings_all,
)




"""
Databank imports
"""


#[
from __future__ import annotations

from typing import (Self, )
from collections.abc import (Iterable, Callable, )
import csv as _cs
import numpy as _np
import dataclasses as _dc
import pickle as _pickle

from ..series import main as _series
from .. import dates as _dates
#]


@_dc.dataclass
class _ImportBlockDescriptor:
    """
    """
    #[
    row_index: Iterable[int] | None = None,
    dates: Iterable[_dates.Dater] | None = None,
    column_start: int | None = None,
    num_columns: int | None = None,
    names: Iterable[str] | None = None,
    descriptions: Iterable[str] | None = None,

    def column_iterator(self, /, ):
        """
        """
        status = False
        names = self.names + [""]
        descriptions = self.descriptions + [""]
        current_columns = None
        current_name = None
        current_description = None
        for i, (n, d) in enumerate(zip(names, descriptions, )):
            if status and n!="*":
                status = False
                yield current_columns, current_name, current_description
            if status and n=="*":
                current_columns.append(i, )
            if not status and n and n!="*":
                status = True
                current_columns = [i]
                current_name = n
                current_description = d
    #]


class DatabankImportMixin:
    """
    """
    #[
    @classmethod
    def from_sheet(
        cls,
        file_name: str,
        /,
        start_date_only: bool = False,
        description_row: bool = False,
        delimiter: str = ",",
        csv_reader_settings: dict = {},
        numpy_reader_settings: dict = {},
        **kwargs,
    ) -> Self:
        """
        """
        self = cls(**kwargs)
        num_header_lines = 1 + int(description_row)
        csv_lines = _read_csv(file_name, num_header_lines, **csv_reader_settings, )
        header_lines = csv_lines[0:num_header_lines]
        data_lines = csv_lines[num_header_lines:]
        name_line = header_lines[0]
        description_row = header_lines[1] if description_row else [""] * len(name_line)
        #
        for b in _block_iterator(name_line, description_row, data_lines, start_date_only, ):
            array = _read_array_for_block(file_name, b, num_header_lines, delimiter=delimiter, **numpy_reader_settings, )
            _add_series_for_block(self, b, array, )
        #
        return self

    @classmethod
    def from_pickle(
        cls,
        file_name: str,
        /,
        **kwargs,
    ) -> Self:
        """
        """
        with open(file_name, "rb") as fid:
            return _pickle.load(fid, **kwargs, )
    #]


def _read_csv(file_name, num_header_lines, /, delimiter=",", **kwargs, ):
    """
    Read CSV cells into a list of lists
    """
    #[
    with open(file_name, "rt", encoding="utf-8-sig", ) as fid:
        all_lines = [ line for line in _cs.reader(fid, **kwargs, ) ]
    return all_lines
    #]


def _remove_nonascii_from_start(string, /, ):
    """
    Remove non-ascii characters from the start of a string
    """
    #[
    while string and not string[0].isascii():
        string = string[1:]
    return string
    #]


def _block_iterator(name_line, description_row, data_lines, start_date_only, /, ):
    """
    """
    #[
    _is_end = lambda cell: cell.startswith("__")
    _is_start = lambda cell: cell.startswith("__") and _dates.Frequency.from_letter(cell) is not _dates.Frequency.UNKNOWN
    name_line += ["__"]
    status = False
    blocks = []
    current_date_column = None
    current_start = None
    current_frequency = None
    num_columns = len(name_line)
    for column, cell in enumerate(name_line):
        if status and _is_end(cell):
            status = False
            num_columns = column - current_start
            names = name_line[current_start:column]
            descriptions = description_row[current_start:column]
            row_index, dates = _extract_dates_from_data_lines(data_lines, current_frequency, current_date_column, start_date_only, )
            yield _ImportBlockDescriptor(row_index, dates, current_start, num_columns, names, descriptions)
        if not status and _is_start(cell):
            status = True
            current_date_column = column
            current_start = column + 1
            current_frequency = _dates.Frequency.from_letter(cell)
    #]


def _extract_dates_from_data_lines(
    data_lines,
    frequency,
    column,
    start_date_only,
    /,
) -> tuple[tuple[int], tuple[_dates.Dater]]:
    """
    """
    #[
    start_date = _dates.Dater.from_sdmx_string(frequency, data_lines[0][column])
    date_extractor = {
        True: lambda i, line: start_date + i,
        False: lambda i, line: _dates.Dater.from_sdmx_string(frequency, line[column]),
    }[start_date_only]
    row_index_and_dates = [ 
        (i, date_extractor(i, line))
        for i, line in enumerate(data_lines)
        if start_date_only or line[column]
    ]
    return tuple(zip(*row_index_and_dates))
    #]


def _read_array_for_block(file_name, block, num_header_lines, /, delimiter=",", **kwargs, ):
    #[
    skip_header = num_header_lines
    usecols = [ c for c in range(block.column_start, block.column_start+block.num_columns) ]
    return _np.genfromtxt(file_name, skip_header=skip_header, usecols=usecols, delimiter=delimiter, ndmin=2, **kwargs)
    #]


def _add_series_for_block(self, block, array, /, ):
    """
    """
    #[
    array = array[block.row_index, :]
    for columns, name, description in block.column_iterator():
        series = _series.Series(num_columns=len(columns), description=description)
        series.set_data(block.dates, array[:, columns])
        self[name] = series
    #]




"""
"""


#[
from __future__ import annotations

from typing import (Self, Callable, )
import copy as _cp

from .. import equations as _equations
from .. import quantities as _quantities
from .. import wrongdoings as _wrongdoings
from ..fords import descriptors as _descriptors
from ..equators import plain as _equators

from . import _flags
#]


_PLAIN_EQUATOR_EQUATION = (
    _equations.EquationKind.TRANSITION_EQUATION
    | _equations.EquationKind.MEASUREMENT_EQUATION
)


class Invariant:
    """
    Invariant part of a Model object
    """
    #[
    __slots__ = (
        "quantities",
        "dynamic_equations",
        "steady_equations",
        "preprocessor",
        "postprocessor",
        "_flags",
        "_function_context",
        "shock_qid_to_std_qid",
        "dynamic_descriptor",
        "steady_descriptor",
        "_plain_equator_for_dynamic_equations",
        "_plain_equator_for_steady_equations",
        "_min_shift",
        "_max_shift",
    )
    def __init__(
        self,
        source,
        /,
        context: dict | None = None,
        check_syntax: bool = True,
        **kwargs,
    ) -> Self:
        """
        """
        self._flags = _flags.Flags.from_kwargs(**kwargs, )
        self._populate_function_context(context)
        #
        self.quantities = _cp.deepcopy(source.quantities)
        self.dynamic_equations = _cp.deepcopy(source.dynamic_equations)
        self.steady_equations = _cp.deepcopy(source.steady_equations)
        self.preprocessor = None
        self.postprocessor = None
        #
        _add_stds(self.quantities, _quantities.QuantityKind.TRANSITION_SHOCK, _quantities.QuantityKind.TRANSITION_STD, )
        _add_stds(self.quantities, _quantities.QuantityKind.MEASUREMENT_SHOCK, _quantities.QuantityKind.MEASUREMENT_STD, )
        _quantities.check_unique_names(self.quantities)
        #
        quantities = _quantities.reorder_by_kind(self.quantities, )
        dynamic_equations = _equations.reorder_by_kind(self.dynamic_equations, )
        steady_equations = _equations.reorder_by_kind(self.steady_equations, )
        _quantities.stamp_id(self.quantities, )
        _equations.stamp_id(self.dynamic_equations, )
        _equations.stamp_id(self.steady_equations, )
        #
        self.shock_qid_to_std_qid = _create_shock_qid_to_std_qid(self.quantities, )
        #
        name_to_qid = _quantities.create_name_to_qid(self.quantities, )
        _equations.finalize_dynamic_equations(self.dynamic_equations, name_to_qid, )
        _equations.finalize_steady_equations(self.steady_equations, name_to_qid, )
        #
        if check_syntax:
            _check_syntax(self.dynamic_equations, self._function_context, )
            _check_syntax(self.steady_equations, self._function_context, )
        #
        self.dynamic_descriptor = _descriptors.Descriptor(self.dynamic_equations, self.quantities, self._function_context, )
        self.steady_descriptor = _descriptors.Descriptor(self.steady_equations, self.quantities, self._function_context, )
        #
        self._plain_equator_for_dynamic_equations = _equators.PlainEquator(
            _equations.generate_equations_of_kind( self.dynamic_equations, _PLAIN_EQUATOR_EQUATION, ),
            custom_functions=self._function_context,
        )
        #
        self._plain_equator_for_steady_equations = _equators.PlainEquator(
            _equations.generate_equations_of_kind(self.steady_equations, _PLAIN_EQUATOR_EQUATION, ),
            custom_functions=self._function_context,
        )
        #
        self._min_shift, self._max_shift = None, None
        self._populate_min_max_shifts()

    def _populate_min_max_shifts(self, /, ) -> None:
        """
        """
        self._min_shift = _equations.get_min_shift_from_equations(
            self.dynamic_equations + self.steady_equations,
        )
        self._max_shift = _equations.get_max_shift_from_equations(
            self.dynamic_equations + self.steady_equations,
        )

    def _populate_function_context(
        self,
        context: dict | None,
        /,
    ) -> None:
        """
        """
        self._function_context = {
            k: v for k, v in context.items()
            if isinstance(v, Callable)
        } if context else None
    #]


def _check_syntax(equations, function_context, /, ):
    """
    Try all equations at once; if this fails, do equation by equation to # catch the troublemakers
    """
    try:
        eval(_equations.create_equator_func_string(equations), )
    except:
        _catch_troublemakers(equations, function_context, )
    #]


def _catch_troublemakers(equations, function_context, /, ):
    """
    Catch the troublemakers
    """
    #[
    fail = [
        eqn.human for eqn in equations
        if not _success_creating_lambda(eqn, function_context)
    ]
    if fail:
        message = ["Syntax error in these equations"] + fail
        _wrongdoings.throw("error", message)
    #]


def _success_creating_lambda(equation, function_context):
    """
    """
    #[
    try:
        eval(_equations.create_equator_func_string([equation.xtring]))
        return True
    except Exception as ex:
        return False
    #]


def _create_shock_qid_to_std_qid(
    quantities: Iterable[_quantities.Quantity],
    /,
) -> dict[int, int]:
    """
    """
    name_to_qid = _quantities.create_name_to_qid(quantities, )
    qid_to_name = _quantities.create_qid_to_name(quantities, )
    kind = _quantities.QuantityKind.SHOCK
    all_shock_qids = tuple(_quantities.generate_qids_by_kind(quantities, kind))
    return {
        shock_qid: name_to_qid[_create_std_name(qid_to_name[shock_qid], )]
        for shock_qid in all_shock_qids
    }


def _add_stds(
    quantities: Iterable[_quantities.Quantity],
    shock_kind: _quantities.QuantityKind,
    std_kind: _quantities.QuantityKind,
    /,
) -> None:
    """
    """
    shocks = (q for q in quantities if q.kind in shock_kind)
    for std_qid, shock in enumerate(shocks, start=len(quantities)):
        std_human = _create_std_name(shock.human, )
        std_logly = False
        std_description = _create_std_description(shock.description, shock.human, )
        std_entry = len(quantities)
        std_quantity = _quantities.Quantity(std_qid, std_human, std_kind, std_logly, std_description, std_entry, )
        quantities.append(std_quantity, )


_STD_PREFIX = "std_"
_STD_DESCRIPTION = "(Std) "


def _create_std_name(
    shock_name: str,
    /,
) -> str:
    """
    """
    return _STD_PREFIX + shock_name


def _create_std_description(
    shock_description: str,
    shock_human: str,
    /,
) -> str:
    """
    """
    return _STD_DESCRIPTION + (shock_description if shock_description else shock_human)




"""
Main time series class definition
"""


#[
from __future__ import annotations

from numbers import Number
from collections.abc import (Iterable, Callable, )
from typing import (Self, TypeAlias, )
from types import (EllipsisType, )
import numpy as _np
import itertools as _it
import functools as _ft

from ..conveniences import views as _views
from ..conveniences import descriptions as _descriptions
from ..conveniences import copies as _copies
from .. import dates as _dates
from .. import wrongdoings as _wrongdoings

from . import _plotly
from . import _conversion
from . import _x13
from . import _hp
#]


FUNCTION_ADAPTATIONS_NUMPY = {"log", "exp", "sqrt", "maximum", "minimum", "mean", "median"}
FUNCTION_ADAPTATIONS_BUILTINS = {"max", "min", }
FUNCTION_ADAPTATIONS = FUNCTION_ADAPTATIONS_NUMPY | FUNCTION_ADAPTATIONS_BUILTINS


__all__ = [
    "Series", "shift",
    "diff", "difflog", "pct", "roc",
    "cum_diff", "cum_difflog", "cum_pct", "cum_roc",
] + list(FUNCTION_ADAPTATIONS)


Dates: TypeAlias = _dates.Dater | Iterable[_dates.Dater] | _dates.Ranger | EllipsisType | None
ColumnsRequestType: TypeAlias = int | Iterable[int] | slice | None
Data: TypeAlias = Number | Iterable[Number] | tuple | _np.ndarray


def _get_date_positions(dates, base, num_periods):
    pos = list(_dates.date_index(dates, base))
    min_pos = _builtin_min((x for x in pos if x is not None), default=0)
    max_pos = _builtin_max((x for x in pos if x is not None), default=0)
    add_before = _builtin_max(-min_pos, 0)
    add_after = _builtin_max(max_pos - num_periods, 0)
    pos_adjusted = [
        p + add_before if p is not None else None
        for p in pos
    ]
    return pos_adjusted, add_before, add_after


def _trim_decorate(func):
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        self.trim()
        return self
    return wrapper


class Series(
    _hp.Mixin,
    _x13.Mixin,
    _plotly.Mixin,
    _descriptions.DescriptionMixin,
    _conversion.Mixin,
    _views.SeriesViewMixin,
    _copies.CopyMixin,
):
    """
    """
    #[
    __slots__ = (
        "start_date", "data", "data_type", "metadata"
        "_description", "_column_titles",
    )

    _numeric_format: str = "15g"
    _short_str_format: str = ">15"
    _date_str_format: str = ">12"
    _missing = _np.nan
    _missing_str: str = "·"
    _test_missing_period = staticmethod(lambda x: _np.all(_np.isnan(x)))

    def __init__(
        self,
        /,
        *,
        num_columns: int = 1,
        data_type: type = _np.float64,
        description: str = "",
        column_titles: Iterable[str] | None = None,
        **kwargs,
    ) -> None:
        self.start_date = None
        self.data_type = data_type
        self.data = _np.full((0, num_columns), self._missing, dtype=self.data_type)
        self.metadata = {}
        self._description = description
        self._column_titles = column_titles or [""] * num_columns

    def reset(self, /, ) -> None:
        self.__init__(
            num_columns=self.num_columns,
            data_type=self.data_type,
        )

    def _create_periods_of_missing_values(self, num_rows=0):
        return _np.full(
            (num_rows, self.shape[1]),
            self._missing,
            dtype=self.data_type
        )

    @property
    def shape(self):
        return self.data.shape

    @property
    def num_columns(self):
        return self.data.shape[1]

    @property
    def range(self):
        return _dates.Ranger(self.start_date, self.end_date) if self.start_date else []

    @property
    def end_date(self):
        return self.start_date + self.data.shape[0] - 1 if self.start_date else None

    @property
    def frequency(self):
        return (
            self.start_date.frequency
            if self.start_date is not None
            else _dates.Frequency.UNKNOWN
        )

    @classmethod
    def from_dates_and_values(
        cls,
        dates: Iterable[_dates.Dater],
        values: _np.ndarray | Iterable,
        /,
    ) -> Self:
        values = _conform_data(values)
        num_columns = values.shape[1] if hasattr(values, "shape") else 1
        self = cls(num_columns=num_columns)
        self.set_data(dates, values)
        return self

    @classmethod
    def from_start_date_and_values(
        cls,
        start_date: _dates.Dater,
        values: _np.ndarray | Iterable,
        /,
        **kwargs,
    ) -> Self:
        values = _conform_data(values, )
        num_columns = values.shape[1] if hasattr(values, "shape") else 1
        self = cls(num_columns=num_columns, **kwargs)
        self.start_date = start_date
        self.data = values
        self.trim()
        return self

    @classmethod
    def from_dates_and_func(
        cls,
        dates: Iterable[_dates.Dater],
        func: Callable,
        /,
        **kwargs,
    ) -> Self:
        """
        Create a new time series from dates and a function
        """
        self = cls(**kwargs)
        dates = list(dates)
        data = [
            [func() for j in range(self.num_columns)]
            for i in range(len(dates))
        ]
        data = _np.array(data, dtype=self.data_type)
        self.set_data(dates, data)
        return self

    @_trim_decorate
    def set_data(
        self,
        dates: Dates,
        data: Data | Series,
        columns: ColumnsRequestType = None,
        /,
    ) -> Self:
        dates = self._resolve_dates(dates)
        columns = self._resolve_columns(columns)
        if not self.start_date:
            self.start_date = next(iter(dates))
            self.data = self._create_periods_of_missing_values(1)
        pos, add_before, add_after = _get_date_positions(dates, self.start_date, self.shape[0]-1)
        self.data = self._create_expanded_data(add_before, add_after)
        if add_before:
            self.start_date -= add_before
        if hasattr(data, "get_data"):
            data = data.get_data(dates)
        if isinstance(data, _np.ndarray):
            self.data[pos, :] = data
            return
        if not isinstance(data, tuple):
            data = _it.repeat(data)
        for c, d in zip(columns, data):
            self.data[pos, c] = d
        return self

    def _get_data_and_recreate(
        self,
        *args,
    ) -> _np.ndarray:
        """
        """
        data, dates = self.get_data_and_resolved_dates(*args)
        num_columns = data.shape[1]
        new = Series(num_columns=num_columns, data_type=self.data_type)
        new.set_data(dates, data)
        return new

    def get_data_and_resolved_dates(
        self,
        dates: Dates,
        columns: ColumnsRequestType = None,
        /,
    ) -> _np.ndarray:
        """
        """
        dates = [ t for t in self._resolve_dates(dates) ]
        columns = self._resolve_columns(columns)
        base_date = self.start_date
        if not dates or not base_date:
            num_dates = len(set(dates))
            return self._create_periods_of_missing_values(num_dates)[:, columns], dates
        pos, add_before, add_after = _get_date_positions(dates, self.start_date, self.shape[0]-1)
        data = self._create_expanded_data(add_before, add_after)
        if not isinstance(pos, Iterable):
            pos = (pos, )
        return data[_np.ix_(pos, columns)], dates

    def get_data(
        self,
        *args,
    ) -> _np.ndarray:
        return self.get_data_and_resolved_dates(*args)[0]

    def get_data_column(
        self,
        dates: Dates,
        column: Number | None = None,
        /,
    ) -> _np.ndarray:
        """
        """
        column = column if column and column<self.data.shape[1] else 0
        return self.get_data(dates, column)

    def extract_columns(
        self,
        columns,
        /,
    ) -> None:
        if not isinstance(columns, Iterable):
            columns = (columns, )
        columns = [ c for c in columns ]
        self.data = self.data[:, columns]
        self.column_titles = [ self.column_titles[c] for c in columns ]

    def set_start_date(
        self,
        new_start_date: _dates.Dater,
        /,
    ) -> Self:
        self.start_date = new_start_date
        return self

    def _resolve_dates(self, dates):
        if dates is None:
            return []
        if dates is Ellipsis:
            dates = _dates.Ranger(None, None)
        if isinstance(dates, _dates.ResolvableProtocol) and dates.needs_resolve:
            dates = dates.resolve(self)
        return dates

    def _resolve_columns(
        self,
        columns: ColumnsRequestType,
        /,
    ) -> Iterable[int]:
        """
        Resolve column request to an iterable of integers
        """
        if columns is None or columns is Ellipsis:
            columns = slice(None)
        if isinstance(columns, slice):
            columns = range(*columns.indices(self.num_columns))
        if not isinstance(columns, Iterable):
            columns = (columns, )
        return columns

    def __call__(self, *args):
        """
        Get data self[dates] or self[dates, columns]
        """
        return self.get_data(*args)

    def _shift(
        self,
        shift_by: int | str,
    ) -> None:
        match shift_by:
            case "yoy":
                self._shift_by_number(-self.frequency.value, )
            case "boy" | "soy":
                self._shift_to_soy()
            case "eopy":
                self._shift_to_eopy()
            case "tty":
                self._shift_to_tty()
            case _:
                self._shift_by_number(shift_by, )

    def _shift_by_number(self, shift_by: int, ) -> None:
        """
        Shift (lag, lead) start date by a number of periods
        """
        self.start_date = (
            self.start_date - shift_by
            if self.start_date else self.start_date
        )

    def _shift_to_soy(self, ) -> None:
        new_data = self.get_data(
            t.create_soy()
            for t in self.range
        )
        self._replace_data(new_data, )

    def _shift_to_tty(self, ) -> None:
        new_data = self.get_data(
            t.create_tty()
            for t in self.range
        )
        self._replace_data(new_data, )

    def _shift_to_eopy(self, ) -> Self:
        new_data = self.get_data(
            t.create_eopy()
            for t in self.range
        )
        self._replace_data(new_data, )

    def __getitem__(self, index):
        """
        Create a new time series based on date retrieved by self[dates] or self[dates, columns]
        """
        if isinstance(index, int):
            new = self.copy()
            new._shift(index)
            return new
        if not isinstance(index, tuple):
            index = (index, None, )
        return self._get_data_and_recreate(*index)

    def __setitem__(self, index, data):
        """
        Set data self[dates] = ... or self[dates, columns] = ...
        """
        if not isinstance(index, tuple):
            index = (index, None, )
        return self.set_data(index[0], data, index[1])

    def hstack(self, *args):
        if not args:
            return self.copy()
        encompassing_range = _dates.get_encompassing_range(self, *args)
        new_data = self.get_data(encompassing_range)
        add_data = (
            x.get_data(encompassing_range) if hasattr(x, "get_data") else _create_data_from_number(x, encompassing_range, self.data_type)
            for x in args
        )
        new_data = _np.hstack((new_data, *add_data))
        new = Series(num_columns=new_data.shape[1])
        new.set_data(encompassing_range, new_data)
        return new

    def clip(
        self,
        new_start_date: _dates.Dater | None,
        new_end_date: _dates.Dater | None,
    ) -> None:
        if new_start_date is None or new_start_date < self.start_date:
            new_start_date = self.start_date
        if new_end_date is None or new_end_date > self.end_date:
            new_end_date = self.end_date
        if new_start_date == self.start_date and new_end_date == self.end_date:
            return
        self.start_date = new_start_date
        self.data = self.get_data(_dates.Ranger(new_start_date, new_end_date))

    def is_empty(self, ) -> bool:
        return not self.data.size

    def empty(self, ) -> None:
        self.data = _np.empty((0, self.num_columns), dtype=self.data_type)

    def overlay_by_range(
        self,
        other: Self,
        /,
    ) -> Self:
        self.trim()
        other.trim()
        self.set_data(other.range, other.data, )
        return self

    def overlay(
        self,
        other: Self,
        /,
        method = "by_range",
    ) -> Self:
        return self._LAY_METHOD_RESOLUTION[method]["overlay"](self, other, )

    def underlay_by_range(
        self,
        other: Self,
        /,
    ) -> Self:
        """
        """
        self.trim()
        other.trim()
        self_range = self.range
        self_data = self.data
        #
        self.start_date = other.start_date
        enforce_columns = _np.zeros((1, self.data.shape[1]), dtype=self.data_type)
        self.data = enforce_columns + other.data
        self.set_data(self_range, self_data, )
        return self

    def underlay(
        self,
        other: Self,
        /,
        method = "by_range",
    ) -> Self:
        return self._LAY_METHOD_RESOLUTION[method]["underlay"](self, other, )

    _LAY_METHOD_RESOLUTION = {
        "by_range": {"overlay": overlay_by_range, "underlay": underlay_by_range},
    }

    def __or__(self, other):
        return self.hstack(other)

    def __lshift__(self, other):
        return self.copy().overlay_by_range(other, )

    def __rshift__(self, other):
        return self.copy().overlay_by_range(self, )

    def trim(self):
        if self.data.size==0:
            self.reset()
            return self
        num_leading = _get_num_leading_missing_rows(self.data, self._test_missing_period)
        if num_leading == self.data.shape[0]:
            self.reset()
            return self
        num_trailing = _get_num_leading_missing_rows(self.data[::-1], self._test_missing_period)
        if not num_leading and not num_trailing:
            return self
        slice_from = num_leading if num_leading else None
        slice_to = -num_trailing if num_trailing else None
        self.data = self.data[slice(slice_from, slice_to), ...]
        if slice_from:
            self.start_date += int(slice_from)
        return self

    def _create_expanded_data(self, add_before, add_after):
        return _np.pad(
            self.data, ((add_before, add_after), (0, 0)),
            mode="constant", constant_values=self._missing
        )

    def _check_data_shape(self, data, /, ):
        if data.shape[1] != self.data.shape[1]:
            raise Exception("Time series data being assigned must preserve the number of columns")

    def __neg__(self):
        """
        -self
        """
        output = self.copy()
        output.data = -output.data
        return output

    def __pos__(self):
        """
        +self
        """
        return self.copy()

    def __add__(self, other) -> Self|_np.ndarray:
        """
        self + other
        """
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x + y)

    def __radd__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__radd__(other))

    def __mul__(self, other) -> Self|_np.ndarray:
        """
        self + other
        """
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x * y)

    def __rmul__(self, other):
        """
        other + self
        """
        return self.apply(lambda data: data.__rmul__(other))

    def __sub__(self, other):
        """
        self - other
        """
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x - y)

    def __rsub__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rsub__(other))

    def __pow__(self, other):
        """
        self ** other
        """
        # FIXME: 1 ** numpy.nan -> 1 !!!!!
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x ** y)

    def __rpow__(self, other):
        """
        other ** self
        """
        return self.apply(lambda data: data.__rpow__(other))

    def __truediv__(self, other):
        """
        self - other
        """
        arg = other.data if isinstance(other, type(self)) else other
        return self._binop(other, lambda x, y: x / y)

    def __rtruediv__(self, other):
        """
        other - self
        """
        return self.apply(lambda data: data.__rtruediv__(other))

    def __floordiv__(self, other):
        """
        self // other
        """
        return self._binop(other, lambda x, y: x // y)

    def __rfloordiv__(self, other):
        """
        other // self
        """
        return self.apply(lambda data: data.__rfloordiv__(other))

    def __mod__(self, other):
        """
        self % other
        """
        return self._binop(other, lambda x, y: x % y)

    def __rmod__(self, other, /, ):
        """
        other % self
        """
        return self.apply(lambda data: data.__rmod__(other))

    def apply(self, func, /, *args, **kwargs, ):
        new_data = func(self.data, *args, **kwargs, )
        axis = kwargs.get("axis", )
        if new_data.shape == self.data.shape:
            new = self.copy()
            new._replace_data(new_data, )
            return new
        elif (axis is None or axis == 1) and new_data.shape == (self.data.shape[0], ):
            new = self.copy()
            new.data = new_data.reshape(self.data.shape[0], 1, )
            new.trim()
            return new
        elif axis is None and new_data.shape == ():
            return new_data
        elif axis == 0 and new_data.shape == (self.data.shape[1], ):
            return new_data
        else:
            raise Exception("Function applied on a time series resulted in a data array with an unexpected shape")

    def _binop(self, other, func, /, ):
        if not isinstance(other, type(self)):
            return self.apply(lambda data: func(data, other))
        encompassing_range = _dates.get_encompassing_range(self, other)
        self_data = self.get_data(encompassing_range)
        other_data = other.get_data(encompassing_range)
        new = Series()
        new.start_date = encompassing_range[0]
        new.data = func(self_data, other_data)
        new.trim()
        return new

    def _replace_data(self, new_data, /, ) -> None:
        self.data = new_data
        self.trim()

    for n in FUNCTION_ADAPTATIONS:
        exec(f"def _{n}_(self, *args, **kwargs, ): return self.apply(_np.{n}, *args, **kwargs, )")
    #]


Series.from_dates_and_data = _wrongdoings.obsolete(Series.from_dates_and_values)
Series.from_start_date_and_data = _wrongdoings.obsolete(Series.from_start_date_and_values)


def _get_num_leading_missing_rows(data, test_missing_period, /, ):
    try:
        num = next(
            i for i, period_data in enumerate(data)
            if not test_missing_period(period_data)
        )
    except StopIteration:
        num = data.shape[0]
    return num


# def apply(x, func: Callable, /, *args, **kwargs) -> object:
    # if isinstance(x, Series):
        # return x.apply(func, *args, **kwargs)
    # else:
        # return func(x, *args, **kwargs)

def hstack(first, *args) -> Self:
    return first.hstack(*args)


def _create_data_from_number(
    number: Number,
    range: _dates.Ranger,
    data_type: type,
    /,
) -> _np.ndarray:
    return _np.full((len(range), 1), number, dtype=data_type)


def shift(x, by=-1, ) -> Series:
    """
    """
    #[
    new = x.copy()
    new._shift(by)
    return new
    #]


def _negative_by_decorator(func):
    def wrapper(x, *args, **kwargs, ):
        if len(args) > 0:
            by = args[0]
            if not isinstance(by, str) and (int(by) != by or by >= 0):
                raise Exception("Time shift must be a negative integer")
        return func(x, *args, **kwargs, )
    return wrapper


@_negative_by_decorator
def diff(x, by=-1, /, ) -> Series:
    return x - shift(x, by)


@_negative_by_decorator
def cum_diff(dx, by=-1, /, initial=0, range=_dates.Ranger(), ) -> Series:
    return _cumulate(dx, by, "diff", initial, range, )


@_negative_by_decorator
def difflog(x, by=-1, /, ) -> Series:
    return log(x) - log(shift(x, by))


@_negative_by_decorator
def cum_difflog(dx, by=-1, /, initial=1, range=_dates.Ranger(), ) -> Series:
    return _cumulate(dx, by, "difflog", initial, range, )


@_negative_by_decorator
def pct(x, by=-1, /, ) -> Series:
    return 100*(x/shift(x, by) - 1)


@_negative_by_decorator
def cum_pct(dx, by=-1, /, initial=1, range=_dates.Ranger(), ) -> Series:
    return _cumulate(dx, by, "pct", initial, range, )


@_negative_by_decorator
def roc(x, by=-1, /, ) -> Series:
    return x/shift(x, by)


@_negative_by_decorator
def cum_roc(roc, by=-1, /, initial=1, range=_dates.Ranger(), ) -> Series | None:
    return _cumulate(roc, by, "roc", initial, range, )


_CUMULATIVE_FUNCTIONS = {
    "diff": {
        "forward": lambda x_past, change_curr: x_past + change_curr,
        "backward": lambda x_future, change_future: x_future - change_future,
    },
    "difflog": {
        "forward": lambda x_past, change_curr: x_past * exp(change_curr),
        "backward": lambda x_future, change_future: x_future / exp(change_future),
    },
    "pct": {
        "forward": lambda x_past, change_curr: x_past * (1 + change_curr/100),
        "backward": lambda x_future, change_future: x_future / (1 + change_future/100),
    },
    "roc": {
        "forward": lambda x_past, change_curr: x_past * change_curr,
        "backward": lambda x_future, change_future: x_future / change_future,
    },
}


def _cumulate(dx, by, func, initial, range, /, ) -> Series:
    """
    """
    #[
    direction = range.direction
    cum_func = _CUMULATIVE_FUNCTIONS[func][direction]
    new = Series(num_columns=dx.num_columns, data_type=dx.data_type, )
    match direction:
        case "forward":
            _cumulate_forward(new, dx, by, cum_func, initial, range, )
        case "backward":
            _cumulate_backward(new, dx, by, cum_func, initial, range, )
    return new.trim()
    #]


def _cumulate_forward(new, dx, by, cum_func, initial, range, /, ) -> None:
    """
    """
    #[
    range = range.resolve(dx)
    shifted_range = tuple(t.shift(by) for t in range)
    initial_range = _dates.Ranger(_builtin_min(shifted_range), range.end_date, )
    new.set_data(initial_range, initial)
    for t, sh in zip(range, shifted_range):
        new_data = cum_func(new.get_data(sh, ), dx.get_data(t, ), )
        new.set_data(t, new_data)
    #]


def _cumulate_backward(new, dx, by, cum_func, initial, shifted_backward_range, /, ) -> None:
    """
    """
    #[
    dx_range_shifted = _dates.Ranger(dx.start_date, dx.end_date, -1, )
    dx_range_shifted.shift(by)
    shifted_backward_range = shifted_backward_range.resolve(dx_range_shifted)
    backward_range = shifted_backward_range.copy()
    backward_range.shift(-by)
    initial_range = _dates.Ranger(_builtin_min(shifted_backward_range), backward_range.start_date)
    new.set_data(initial_range, initial)
    for t, sh in zip(backward_range, shifted_backward_range):
        new_data = cum_func(new.get_data(t, ), dx.get_data(t, ), )
        new.set_data(sh, new_data)
    #]


def _conform_data(data, /, ) -> _np.ndarray:
    """
    """
    #[
    if not isinstance(data, _np.ndarray, ):
        data = _np.array(data, dtype=float, ndmin=2, ).T
    return data
    #]


for n in FUNCTION_ADAPTATIONS_NUMPY:
    exec(
        f"@_ft.singledispatch"
        f"\n"
        f"def {n}(x, *args, **kwargs): "
        f"return _np.{n}(x, *args, **kwargs)"
    )
    exec(
        f"@{n}.register(Series, )"
        f"\n"
        f"def _(x, *args, **kwargs): "
        f"return x._{n}_(*args, **kwargs)"
    )


for n in FUNCTION_ADAPTATIONS_BUILTINS:
    exec(
        f"_builtin_{n} = {n}"
    )
    exec(
        f"@_ft.singledispatch"
        f"\n"
        f"def {n}(x, *args, **kwargs): "
        f"return _builtin_{n}(x, *args, **kwargs)"
    )
    exec(
        f"@{n}.register(Series, )"
        f"\n"
        f"def {n}(x, *args, **kwargs): "
        f"return x._{n}_(*args, **kwargs)"
    )




"""
"""


#[
from __future__ import annotations
from typing import TYPE_CHECKING

from numbers import (Number, )
from collections.abc import (Iterable, Callable, )
from typing import (TypeAlias, Any, Literal, )
import copy as _copy
import functools as _ft
import itertools as _it
import numpy as _np
import scipy as _sp

from .. import equations as _eq
from .. import quantities as _quantities
from .. import wrongdoings as _wrongdoings
from ..fords import steadiers as _fs
from ..evaluators import steady as _evaluators

from . import variants as _variants
from . import _flags
#]


class SteadyMixin:
    """
    """
    def steady(
        self,
        /,
        **kwargs,
    ) -> list[dict]:
        """
        Calculate steady state for each Variant within this Model
        """
        model_flags = _flags.Flags.update_from_kwargs(self.get_flags(), **kwargs)
        solver = self._choose_steady_solver(model_flags.is_linear, model_flags.is_flat, )
        info = tuple(
            solver(v, model_flags, **kwargs, )
            for v in self._variants
        )
        return info

    def _steady_linear(
        self, 
        variant: _variants.Variant,
        model_flags: _flags.Flags,
        /,
        *,
        algorithm: Callable,
        **kwargs,
    ) -> dict[str, Any]:
        """
        """
        #
        # Calculate first-order system for steady equations for this variant
        system = self._systemize(
            variant,
            self._invariant.steady_descriptor,
            model_flags,
        )
        #
        #=======================================================================
        # Core algorithm: Calculate steady state for this variant
        Xi, Y, dXi, dY = algorithm(system)
        levels = _np.hstack(( Xi.flat, Y.flat )).flatten()
        changes = _np.hstack(( dXi.flat, dY.flat )).flatten()
        #=======================================================================
        #
        # From first-order state space, extract only tokens with zero shift
        tokens = tuple(_it.chain(
            self._invariant.steady_descriptor.system_vectors.transition_variables,
            self._invariant.steady_descriptor.system_vectors.measurement_variables,
        ))
        #
        # True for tokens with zero shift, e.g. [True, False, True, ... ]
        zero_shift_index = [ not t.shift for t in tokens ]
        #
        # Extract steady levels for quantities with zero shift
        levels = levels[zero_shift_index]
        changes = changes[zero_shift_index]
        qids = tuple(t.qid for t in _it.compress(tokens, zero_shift_index))
        #
        # Delogarithmize when needed
        qid_to_logly = self.create_qid_to_logly()
        levels = _apply_delog_on_vector(levels, qids, qid_to_logly)
        changes = _apply_delog_on_vector(changes, qids, qid_to_logly)
        #
        v.update_levels_from_array(levels, qids, )
        v.update_changes_from_array(changes, qids, )
        #
        info = {}
        return info

    _steady_linear_flat = _ft.partialmethod(
        _steady_linear,
        algorithm=_fs.solve_steady_linear_flat,
    )

    _steady_linear_nonflat = _ft.partialmethod(
        _steady_linear,
        algorithm=_fs.solve_steady_linear_nonflat,
    )

    def _steady_nonlinear(
        self,
        variant: _variants.Variant,
        model_flags: _flags.Flags,
        /,
        evaluator_class: type,
        *,
        fix: Iterable[str] | None = None,
        fix_level: Iterable[str] | None = None,
        fix_change: Iterable[str] | None = None,
        optim_settings: dict[str, Any] | None = None,
        iter_printer_settings: dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        """
        """
        optim_settings = \
            _DEFAULT_OPTIM_SETTINGS if optim_settings is None \
            else _DEFAULT_OPTIM_SETTINGS | optim_settings
        #
        # REFACTOR
        #
        wrt_equations = self.get_steady_equations()
        fixable_quantities = \
            self.get_quantities(kind=_quantities.QuantityKind.ENDOGENOUS_VARIABLE, )
        wrt_qids_levels = \
            _resolve_qids_fixed(fixable_quantities, fix, fix_level, )
        wrt_qids_changes = \
            _resolve_qids_fixed(fixable_quantities, fix, fix_change, )
        steady_evaluator = evaluator_class(
            wrt_equations,
            self.get_quantities(),
            wrt_qids_levels,
            wrt_qids_changes,
            variant,
            iter_printer_settings=iter_printer_settings,
        )
        #
        optim_result = _sp.optimize.root(
           steady_evaluator.eval,
           steady_evaluator.init_guess,
           method="lm",
           jac=True,
           options=optim_settings,
        )
        #
        steady_evaluator.iter_printer.print_footer()
        #
        final_guess = optim_result.x
        func_norm = _sp.linalg.norm(optim_result.fun, 2, )
        success = func_norm < optim_settings["ftol"]
        if not success:
            raise _wrongdoings.IrisPieError(
                "Steady state calculations failed to converge"
            )
        #
        levels, wrt_qids_levels = steady_evaluator.extract_levels(final_guess, )
        variant.update_levels_from_array(levels, wrt_qids_levels, )
        changes, wrt_qids_changes = steady_evaluator.extract_changes(final_guess, )
        variant.update_changes_from_array(changes, wrt_qids_changes, )
        #
        info = {
            "success": success,
            "status": optim_result.status,
            "num_func_evals": optim_result.nfev,
            "num_jacob_evals": optim_result.njev,
            "optim_result": optim_result,
        }
        return info

    _steady_nonlinear_flat = _ft.partialmethod(
        _steady_nonlinear,
        evaluator_class=_evaluators.FlatSteadyEvaluator,
    )

    _steady_nonlinear_nonflat = _ft.partialmethod(
        _steady_nonlinear,
        evaluator_class=_evaluators.NonflatSteadyEvaluator,
    )

    def _choose_steady_solver(
        self,
        is_linear: bool,
        is_flat: bool,
        /,
    ) -> Callable:
        """
        Choose steady solver depending on linear and flat flags
        """
        match (is_linear, is_flat):
            case (False, False):
                return self._steady_nonlinear_nonflat
            case (False, True):
                return self._steady_nonlinear_flat
            case (True, False):
                return self._steady_linear_nonflat
            case (True, True):
                return self._steady_linear_flat

    def check_steady(
        self,
        /,
        equation_switch: Literal["steady", "dynamic"] = "dynamic",
        when_fails: _wrongdoings.HOW = "error",
        tolerance: float = 1e-12,
    ) -> tuple[bool, tuple[dict, ...]]:
        """
        Verify currently assigned steady state in dynamic or steady equations for each Variant within this Model
        """
        qid_to_logly = self.create_qid_to_logly()
        equator = self._choose_plain_equator(equation_switch)
        steady_arrays = (
            v.create_steady_array(
                qid_to_logly,
                num_columns=equator.min_num_columns + 1,
                shift_in_first_column=equator.min_shift,
            ) for v in self._variants
        )
        #
        # REFACTOR
        #
        t_zero = -equator.min_shift
        dis = [ 
            _np.hstack((
                equator.eval(x, t_zero, x[:, t_zero]),
                equator.eval(x, t_zero+1, x[:, t_zero+1]),
            ))
            for x in steady_arrays
        ]
        #
        # REFACTOR
        #
        max_abs_dis = [ _np.max(_np.abs(d)) for d in dis ]
        status = [ d < tolerance for d in max_abs_dis ]
        all_status = all(status)
        if not all_status:
            message = "Invalid steady state"
            _wrongdoings.throw(when_fails, message)
        details = tuple(
            {"discrepancies": d, "max_abs_discrepancy": m, "is_valid": s}
            for d, m, s in zip(dis, max_abs_dis, status)
        )
        return all_status, details


def _apply_delog_on_vector(
    vector: _np.ndarray,
    qids: Iterable[int],
    qid_to_logly: dict[int, bool],
    /,
) -> _np.ndarray:
    """
    Delogarithmize the elements of numpy vector that have True log-status
    """
    logly_index = [ qid_to_logly[qid] for qid in qids ]
    vector = _np.copy(vector)
    if any(logly_index):
        vector[logly_index] = _np.exp(vector[logly_index])
    return vector

def _resolve_qids_fixed(
    fixable_quantities,
    fix: Iterable[str] | str | None,
    fix_spec: Iterable[str] | str | None,
) -> tuple[int, ...]:
    """
    """
    if fix is None:
        fix = ()
    elif isinstance(fix, str):
        fix = (fix, )
    if fix_spec is None:
        fix_spec = ()
    elif isinstance(fix_spec, str):
        fix_spec = (fix_spec, )
    fix = tuple(i for i in set(fix) | set(fix_spec) if not i.startswith("!"))
    qids_fixed, invalid_names = \
        _quantities.lookup_qids_by_name(fixable_quantities, fix, )
    if invalid_names:
        raise _wrongdoings.IrisPieError(
            (f"Cannot fix these names:", ) + invalid_names
        )
    wrt_qids = tuple(
        q.id
        for q in fixable_quantities
        if q.id not in qids_fixed
    )
    return wrt_qids


_DEFAULT_OPTIM_SETTINGS = {
    "ftol": 1e-12,
    "xtol": 1e-12,
    "gtol": 1e-12,
}




"""
"""


#[
from __future__ import annotations

from ..sequentials import main as _sequentials
#]


class ProcessorMixin:
    """
    """
    #[
    def set_preprocessor(
        self,
        preprocessor: _sequentials.Sequential,
        /,
    ) -> None:
        self._invariant.preprocessor = preprocessor

    def set_postprocessor(
        self,
        postprocessor: _sequentials.Sequential,
        /,
    ) -> None:
        self._invariant.postprocessor = postprocessor

    def preprocess(
        self,
        *args,
        **kwargs,
    ) -> tuple:
        return self._invariant.preprocessor.simulate(*args, **kwargs, )

    def postprocess(
        self,
        *args,
        **kwargs,
    ) -> tuple:
        return self._invariant.postprocessor.simulate(*args, **kwargs, )
    #]




"""
Incidence tokens and wrt tokens
"""


#[
from collections.abc import (Iterable, )
from numbers import (Number, )
from typing import (
    NamedTuple, Callable, Self, 
    Protocol, TypeAlias,
)
import operator as _op
import itertools as _it
import numpy as _np

from .. import quantities as _quantities
#]


_PRINT_TOKEN = "x[({qid},t{shift:+g})]"
_PRINT_TOKEN_ZERO_SHIFT = "x[({qid},t)]"


class Token(NamedTuple):
    """
    Incidence
    """
    #[
    qid: int
    shift: int

    def shifted(
        self: Self,
        by: int,
    ) -> Self:
        return Token(self.qid, self.shift+by)

    def print(
        self,
        qid_to_name: dict[int, str], 
    ) -> str:
        s = qid_to_name[self.qid]
        if self.shift:
            s += f"{{{self.shift:+g}}}"
        return s

    def print_xtring(
        self,
    ) -> str:
        return (
            _PRINT_TOKEN.format(qid=self.qid, shift=self.shift)
            if self.shift else _PRINT_TOKEN_ZERO_SHIFT.format(qid=self.qid)
        )
    #]


def get_max_shift(tokens: Iterable[Token]) -> int:
    return max(tok.shift for tok in tokens) if tokens else None


def get_min_shift(tokens: Iterable[Token]) -> int:
    return min(tok.shift for tok in tokens) if tokens else None


def get_max_qid(tokens: Iterable[Token]) -> int:
    return max(tok.qid for tok in tokens) if tokens else None


def generate_qids_from_tokens(tokens: Iterable[Token]) -> Iterable[int]:
    return (tok.qid for tok in tokens)


def get_some_shift_by_quantities(tokens: Iterable[Token], something: Callable) -> dict:
    key = _op.attrgetter("qid")
    sorted_tokens = sorted(tokens, key=key)
    return { k: something(t.shift for t in tokens) for k, tokens in _it.groupby(sorted_tokens, key=key) }


def generate_tokens_of_kinds(tokens: Iterable[Token], qid_to_kind: dict, kinds: _quantities.QuantityKind) -> Iterable[Token]:
    return (tok for tok in tokens if qid_to_kind[tok.qid] in kinds)


def sort_tokens(tokens: Iterable[Token]) -> Iterable[Token]:
    """
    Sort tokens by shift and id
    """
    return sorted(tokens, key=lambda x: (-x.shift, x.qid))


def is_qid_in_tokens(tokens: Iterable[Token], qid: int) -> bool:
    return any(tok.qid == qid for tok in tokens)


def print_tokens(
    tokens: Iterable[Token],
    qid_to_name: dict[int, str]
) -> Iterable[str]:
    """
    Create list of printed tokens
    """
    return [ t.print(qid_to_name) for t in tokens ]




"""
"""

#[
from __future__ import annotations

from typing import (Any, )
import numpy as _np

from ..databanks import main as _databanks
from ..plans import main as _plans
from ..explanatories import main as _explanatories
from .. import dataslabs as _dataslabs
#]


class SimulateMixin:
    """
    """
    #[
    def simulate(
        self,
        in_databank: _databanks.Databank,
        base_range: Iterable[Dater],
        /,
        plan: _plans.Plan | None = None,
        prepend_input: bool = True,
        add_to_databank: _databanks.Databank | None = None,
    ) -> tuple[_databanks.Databank, dict[str, Any]]:
        """
        """
        #
        # Arrange input data into a dataslab
        #
        ds = _dataslabs.Dataslab.from_databank_for_simulation(
            self, in_databank, base_range, column=0, plan=plan,
        )
        #
        # Fill missing residuals with zeros
        #
        ds.fill_missing_in_base_columns(self.res_names, fill=0, )
        #
        # Run simulation
        #
        columns_to_simulate = ds.base_columns
        name_to_row = ds.create_name_to_row()
        self._simulate(ds.data, columns_to_simulate, name_to_row, plan, )
        #
        # Build output databank
        #
        ds.remove_terminal()
        out_db = ds.to_databank()
        if prepend_input:
            out_db.prepend(in_databank, ds.column_dates[0]-1, )
        #
        # Add to custom databank
        #
        if add_to_databank is not None:
            out_db = add_to_databank | out_db
        #
        info = {"dataslab": ds, }
        return out_db, info

    def _simulate(
        self,
        data: _np.ndarray,
        columns: Iterable[int],
        name_to_row: dict[str, int],
        plan: _plans.Plan | None,
        /,
    ) -> None:
        """
        """
        for plan_column, data_column in enumerate(columns):
            for x in self.explanatories:
                status, value = _is_exogenized(x, data, plan, data_column, plan_column, name_to_row, )
                if status:
                    x.exogenize(data, data_column, value)
                else:
                    x.simulate(data, data_column, )

    #]

def _is_exogenized(
    explanatory: _explanatories.Explanatory,
    data: _np.ndarray,
    plan: _plans.Plan | None,
    data_column: int,
    plan_column: int,
    name_to_row: dict[str, int],
) -> bool:
    if plan is None or explanatory.is_identity:
        return False, None
    #
    lhs_name = explanatory.lhs_name
    transform = plan.exogenized[lhs_name][plan_column]
    if transform is None:
        return False, None
    #
    transform_name = transform.resolve_databank_name(lhs_name, )
    lhs_name_row = name_to_row[lhs_name]
    transform_row = name_to_row[transform_name]
    transform_value = data[transform_row, data_column]
    lagged_value = data[lhs_name_row, data_column-1]
    implied_value = transform.eval_exogenized(transform_value, lagged_value, )
    if transform.when_data and _np.isnan(implied_value):
        return False, None
    return True, implied_value




"""
Plotly interface to time series objects
"""


#[
from __future__ import annotations

from types import EllipsisType
import os as _os
import json as _js
import copy as _cp
import plotly.graph_objects as _pg
import plotly.subplots as _ps

from .. import dates as _dates
#]


_PLOTLY_STYLES_FOLDER = _os.path.join(_os.path.dirname(__file__), "plotly_styles")

_PLOTLY_STYLES = {
    "layouts": {},
}
with open(_os.path.join(_PLOTLY_STYLES_FOLDER, "plain_layout.json", ), "rt", ) as fid:
    _PLOTLY_STYLES["layouts"]["plain"] = _js.load(fid, )


builtin_range = range


_COLOR_ORDER = [
  "rgb(  0,   114,   189)",
  "rgb(217,    83,    25)",
  "rgb(237,   177,    32)",
  "rgb(126,    47,   142)",
  "rgb(119,   172,    48)",
  "rgb( 77,   190,   238)",
  "rgb(162,    20,    47)",
]


class Mixin:
    """
    """
    #[
    def plot(
        self,
        *,
        range: Iterable[_dates.Dater] | EllipsisType = ...,
        title: str | None = None,
        legend: Iterable[str] | None = None,
        show_legend: bool | None = None,
        figure = None,
        subplot: tuple[int, int] | int | None = None,
        xline = None,
    ) -> _pg.Figure:
        range = self._resolve_dates(range)
        range = [ t for t in range ]
        data = self.get_data(range)
        num_columns = data.shape[1]
        date_str = [ t.to_plotly_date() for t in range ]
        date_format = range[0].frequency.plotly_format
        figure = _pg.Figure() if figure is None else figure
        axis_id = f"{subplot+1}" if subplot else ""
        subplot = _resolve_subplot(figure, subplot, )
        show_legend = show_legend if show_legend is not None else legend is not None
        for i in builtin_range(num_columns):
            figure.add_trace(
                _pg.Scatter(
                    x=date_str, y=data[:, i], name=legend[i] if legend else None,
                    line={"color": _COLOR_ORDER[i % len(_COLOR_ORDER)]},
                    #marker={"color": _COLOR_ORDER[i % len(_COLOR_ORDER)]},
                    showlegend=show_legend, xhoverformat="%Y-%q",
                    xaxis=f"x{axis_id}", yaxis=f"y{axis_id}",
                ),
                row=subplot[0], col=subplot[1]
            )

        layout = _cp.deepcopy(_PLOTLY_STYLES["layouts"]["plain"])
        layout[f"xaxis{axis_id}"] = layout["xaxis"]
        layout[f"yaxis{axis_id}"] = layout["yaxis"]
        if axis_id:
            del layout["xaxis"]
            del layout["yaxis"]
        layout[f"yaxis{axis_id}"]["tickformat"] = date_format
        layout["title"]["text" ] = title
        figure.update_layout(layout, )

        if xline:
            xline = xline.to_plotly_date()
            figure.add_vline(xline)
        return figure
    #]


def _resolve_subplot(
    figure: _pg.Figure,
    subplot: tuple[int, int] | int | None,
    /,
) -> tuple[int, int]:
    """
    """
    if subplot is None:
        return None, None
    if isinstance(subplot, tuple) or isinstance(subplot, list):
        return subplot[0], subplot[1]
    if isinstance(subplot, int):
        position = subplot
        rows, columns = figure._get_subplot_rows_columns()
        num_rows = len(rows)
        num_columns = len(columns)
        row = position // num_columns + 1
        column = position % num_columns + 1
        return row, column




"""
Resolve pseudofunctions in model source string
"""


#[
from __future__ import annotations

import re as re_

from ..parsers import (shifts as sh_, )
#]


#••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••
# Exposure
#••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••


def resolve_pseudofunctions(source: str, /, ) -> str:
    """
    Substitute expanded pseudofunctions in source string
    """
    return re_.sub(_PSEUDOFUNC_PATTERN, _expand_pseudofunction, source, )


#••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••
# Implementation
#••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••••


def _expand_pseudofunction(match, /, ) -> str:
    func_name, expression, shift = match.group(1), match.group(2), match.group(3)
    func, default_shift = _PSEUDOFUNC_RESOLUTION[func_name]
    shift = _resolve_shift(shift, default_shift)
    return func(expression, shift)


def _resolve_shift(shift: str, default_shift: int, /, ) -> int:
    shift = shift.strip() if shift else ""
    return int(shift) if shift else default_shift


_NAME_MAYBE_WITH_SHIFT = re_.compile(
    # Name starting with a letter
    r"(\b[a-zA-Z]\w*\b)" 
    # Optional square bracketed time shift, capture only the shift
    # (part of sh.TIME_SHIFT_INSIDE)
    r"(?:\[(" + sh_.TIME_SHIFT_INSIDE + r")\])?" 
    # Not followed by ( or [
    r"(?![\(\[])" 
)


def _shift_all_names(
    source: str,
    by: int,
    /,
) -> str:
    """
    Shift all names in a source string by a given number of time steps
    """
    def _replace(match, /, ) -> str:
        name, shift = match.group(1), match.group(2)
        shift = eval(shift) if shift else 0
        shift += int(by) if by else 0
        return (name + "[" + str(shift) + "]") if shift else name
    return re_.sub(_NAME_MAYBE_WITH_SHIFT, _replace, source, )


def _pseudo_shift(code, shift, /, ) -> str:
    return _shift_all_names(code, shift)


def _pseudo_diff(code, shift, /, ) -> str:
    return "(" + "(" + code + ")-(" + _shift_all_names(code, shift) + ")" + ")"


def _pseudo_difflog(code, shift, /, ) -> str:
    return "(" + "log(" + code + ")-log(" + _shift_all_names(code, shift) + ")" + ")"


def _pseudo_pct(code, shift, /, ) -> str:
    return "(" + "100*(" + code + ")/(" + _shift_all_names(code, shift) + ")-100" + ")"


def _pseudo_roc(code, shift, /, ) -> str:
    return "(" + "(" + code + ")/(" + _shift_all_names(code, shift) + ")" + ")"


def _pseudo_mov(code, shift, /, ) -> list[str]:
    if shift == 0:
        return ["0"], 0
    elif shift == 1 or shift == -1:
        return [code], 1
    else:
        step, total = (1, shift) if shift > 0 else (-1, -shift)
        return [
            "(" + _shift_all_names(code, sh) + ")" 
            for sh in range(0, shift, step)
        ], total


def _pseudo_movsum(code, shift, /, ) -> str:
    sequence, total = _pseudo_mov(code, shift, )
    return "(" + "+".join(sequence) + ")"


def _pseudo_movavg(code, shift, /, ) -> str:
    sequence, total = _pseudo_mov(code, shift, )
    return "(" + "(" + "+".join(sequence) + ")/" + str(total) + ")"


def _pseudo_movprod(code, shift, /, ) -> str:
    sequence, *_ = _pseudo_mov(code, shift)
    return "(" + "*".join(sequence) + ")"


_PSEUDOFUNC_RESOLUTION = {
    "shift":
        (_pseudo_shift, -1),
    "diff":
        (_pseudo_diff, -1),
    "difflog":
        (_pseudo_difflog, -1),
    "pct": 
        (_pseudo_pct, -1),
    "roc": 
        (_pseudo_roc, -1),
    "movsum":
        (_pseudo_movsum, -4),
    "movavg":
        (_pseudo_movavg, -4),
    "movprod":
        (_pseudo_movprod, -4),
}


# One of the pseudofunction names as a whole word
_PSEUDOFUNC_NAME_PATTERN = r"\b(" + "|".join(_PSEUDOFUNC_RESOLUTION.keys()) + r")\b"

# Groupped text with no comma and possibly ONE LEVEL of parentheses
_PSEUDOFUNC_EXPRESSION_PATTERN = r"((?:[^\(\),]+|\([^\(\),]+\))+)"

# Groupped text with no comma and no parentheses, optional
_PSEUDOFUNC_TIME_SHIFT_PATTERN = r"(?:,([^\(\),]+))?"

# Parenthesized arguments into pseudofunction
_PSEUDOFUNC_BODY_PATTERN = r"\(" + _PSEUDOFUNC_EXPRESSION_PATTERN + _PSEUDOFUNC_TIME_SHIFT_PATTERN+ r"\)"

# Full pseudofunction pattern
_PSEUDOFUNC_PATTERN = re_.compile(_PSEUDOFUNC_NAME_PATTERN + _PSEUDOFUNC_BODY_PATTERN)




"""
"""


#[
from __future__ import annotations

import numpy as np_
import scipy as sp_

from ..fords import (systems as sy_, )
#]


def solve_steady_linear_flat(
    sys: sy_.System,
    /,
) -> tuple[np_.ndarray, np_.ndarray, np_.ndarray, np_.ndarray]:
    """
    """
    #[
    """
    """
    pinv = np_.linalg.pinv
    lstsq = sp_.linalg.lstsq
    vstack = np_.vstack
    hstack = np_.hstack
    A, B, C, F, G, H = sys.A, sys.B, sys.C, sys.F, sys.G, sys.H
    #
    # A @ Xi + B @ Xi{-1} + C = 0
    # F @ Y + G @ Xi + H = 0
    #
    # Xi = -pinv(A + B) @ C
    Xi, *_ = lstsq(-(A + B), C)
    dXi = np_.zeros(Xi.shape)
    #
    # Y = -pinv(F) @ (G @ Xi + H)
    Y, *_ = lstsq(-F, G @ Xi + H)
    dY = np_.zeros(Y.shape)
    #
    return Xi, Y, dXi, dY
    #]


def solve_steady_linear_nonflat(
    sys: sy_.System,
    /,
) -> tuple[np_.ndarray, np_.ndarray, np_.ndarray, np_.ndarray]:
    """
    """
    #[
    """
    """
    # pinv = np_.linalg.pinv
    lstsq = np_.linalg.lstsq
    vstack = np_.vstack
    hstack = np_.hstack
    A, B, C, F, G, H = sys.A, sys.B, sys.C, sys.F, sys.G, sys.H
    num_y = F.shape[0]
    k = 1
    #
    # A @ Xi + B @ Xi{-1} + C = 0:
    # -->
    # A @ Xi + B @ (Xi - dXi) + C = 0
    # A @ (Xi + k*dXi) + B @ (Xi + (k-1)*dXi) + C = 0
    #
    AB = vstack((
        hstack(( A + B, 0*A + (0-1)*B )),
        hstack(( A + B, k*A + (k-1)*B )),
    ))
    CC = vstack((
        C,
        C,
    ))
    # Xi_dXi = -pinv(AB) @ CC
    Xi_dXi, *_ = lstsq(-AB, CC, rcond=None)
    #
    # F @ Y + G @ Xi + H = 0:
    # -->
    # F @ Y + G @ Xi + H = 0
    # F @ (Y + k*dY) + G @ (Xi + k*dXi) + H = 0
    #
    FF = vstack((
        hstack(( F, 0*F )),
        hstack(( F, k*F )),
    ))
    GG = vstack((
        hstack(( G, 0*G )),
        hstack(( G, k*G )),
    ))
    HH = vstack((
        H,
        H,
    ))
    # Y_dY = -pinv(FF) @ (GG @ Xi_dXi + HH)
    Y_dY, *_ = lstsq(-FF, GG @ Xi_dXi + HH, rcond=None)
    #
    # Separate levels and changes
    #
    num_xi = A.shape[1]
    num_y = F.shape[1]
    Xi, dXi = (
        Xi_dXi[0:num_xi, ...],
        Xi_dXi[num_xi:, ...],
    )
    Y, dY = (
        Y_dY[0:num_y, ...],
        Y_dY[num_y:, ...]
    )
    #
    return Xi, Y, dXi, dY
    #]





"""
Sequential models
"""


#[
from __future__ import annotations

from collections.abc import (Iterable, )
from typing import (Any, Callable, )
import numpy as _np

from .. import equations as _equations
from .. import quantities as _quantities
from .. import sources as _sources
from ..conveniences import copies as _copies
from ..conveniences import descriptions as _descriptions
from ..incidences import main as _incidences
from ..incidences import blazer as _blazer
from ..explanatories import main as explanatories
from ..plans import main as _plans

from . import _simulate as _simulate
#]


__all__ = (
    "Sequential",
)


class Sequential(
    _simulate.SimulateMixin,
    _sources.SourceMixin,
    _copies.CopyMixin,
    _descriptions.DescriptionMixin,
):
    """
    """
    #[
    __slots__ = (
        "explanatories",
        "all_names",
        "_description",
    )

    def __init__(
        self,
        /,
    ) -> None:
        self.explanatories = ()
        self.all_names = ()
        self._description = ""

    @classmethod
    def from_equations(
        cls,
        equations: Iterable[_equations.Equation],
        /,
        custom_functions: dict[str, Callable] | None = None,
    ) -> Self:
        """
        """
        self = cls()
        self.explanatories = tuple(
            explanatories.Explanatory(e, custom_functions=custom_functions, )
            for e in equations
        )
        self._collect_all_names()
        self._finalize_explanatories()
        return self

    @classmethod
    def from_source(
        cls,
        source: _sources.AlgebraicSource,
        /,
        context: dict[str, Callable] | None = None,
    ) -> Self:
        """
        """
        self = cls.from_equations(source.dynamic_equations, custom_functions=context, )
        return self

    @property
    def lhs_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of LHS variables in order of appearance.
        """
        return tuple(
            x.lhs_name
            for x in self.explanatories
        )

    @property
    def res_names(self, /, ) -> tuple[str]:
        """
        Tuple of names of LHS variables in order of appearance.
        """
        return tuple(
            x.res_name
            for x in self.explanatories
            if x.res_name is not None
        )

    @property
    def rhs_only_names(self, /, ) -> set[str]:
        """
        Set of names of RHS variables no appearing on the LHS.
        """
        lhs_res_names = self.lhs_names + self.res_names
        return tuple(
            n for n in self.all_names
            if n not in lhs_res_names
        )

    @property
    def identity_index(self, /, ) -> tuple[int]:
        """
        Tuple of indices of identities
        """
        return tuple(
            i for i, x in enumerate(self.explanatories)
            if x.is_identity
        )

    @property
    def nonidentity_index(self, /, ) -> tuple[int]:
        """
        Tuple of indices of nonidentities
        """
        return tuple(
            i for i, x in enumerate(self.explanatories)
            if not x.is_identity
        )

    @property
    def equations(self, /, ) -> tuple[_equations.Equation]:
        """
        Tuple of equations in order of appearance
        """
        return tuple(
            x.equation
            for x in self.explanatories
        )

    @property
    def lhs_quantities(self, /, ) -> tuple[_quantities.Quantity]:
        """
        Tuple of LHS quantities in order of appearance
        """
        lhs_names = self.lhs_names
        kind = _quantities.QuantityKind.LHS_VARIABLE
        logly = False
        return tuple(
            _quantities.Quantity(qid, name, kind, logly, desc, )
            for (qid, name), desc in zip(enumerate(self.all_names), self.descriptions)
            if name in lhs_names
        )

    @property
    def num_equations(self, /, ) -> int:
        """
        Number of equations
        """
        return len(self.equations)

    @property
    def descriptions(self, /, ) -> tuple[str]:
        """
        """
        return tuple(
            x.equation.description
            for x in self.explanatories
        )

    @property
    def incidence_matrix(self, /, ) -> _np.ndarray:
        def _shift_test(tok: _incidences.Token) -> bool:
            return tok.shift == 0
        return _equations.create_incidence_matrix(
            self.equations,
            self.lhs_quantities,
            shift_test=_shift_test,
        )

    @property
    def min_shift(self, /, ) -> int:
        """
        """
        return min(
            x.min_shift
            for x in self.explanatories
        )

    @property
    def max_shift(self, /, ) -> int:
        """
        """
        return max(
            x.max_shift
            for x in self.explanatories
        )

    def reorder_equations(
        self,
        order = Iterable[int],
    ) -> None:
        self.explanatories = [
            self.explanatories[i]
            for i in order
        ]
        self._collect_all_names()
        self._finalize_explanatories()

    def sequentialize(
        self,
        /,
    ) -> tuple[int, ...]:
        """
        Reorder the model equations so that they can be solved sequentially
        """
        if self.is_sequential:
            return tuple(range(self.num_equations))
        eids_reordered = _blazer.sequentialize_strictly(self.incidence_matrix, )
        self.reorder_equations(eids_reordered, )
        return tuple(eids_reordered)

    @property
    def is_sequential(
        self,
        /,
    ) -> bool:
        """
        """
        return _blazer.is_sequential(self.incidence_matrix, )

    def print_equations(
        self,
        /,
        indent: int = 4,
        descriptions: bool = True,
        separator: str = "\n\n",
    ) -> str:
        """
        """
        return separator.join(
            x.print_equation(indent=indent, description=descriptions, )
            for x in self.explanatories
        )

    def _collect_all_names(
        self,
        /,
    ) -> None:
        """
        """
        all_names = set()
        for x in self.explanatories:
            all_names.update(x.all_names)
        lhs_res_names = self.lhs_names + self.res_names
        rhs_only_names = tuple(all_names.difference(lhs_res_names, ))
        self.all_names = self.lhs_names + rhs_only_names + self.res_names

    def create_qid_to_name(self, ) -> dict[int, str]:
        """
        """
        return {
            i: name
            for i, name in enumerate(self.all_names)
        }

    def create_name_to_qid(self, ) -> dict[str, int]:
        """
        """
        return {
            name: i
            for i, name in enumerate(self.all_names)
        }

    def _finalize_explanatories(
        self,
        /,
    ) -> None:
        """
        """
        name_to_qid = self.create_name_to_qid()
        for x in self.explanatories:
            x.finalize(name_to_qid, )

    def __str__(self, /, ) -> str:
        """
        """
        indented = " " * 4
        return "\n".join((
            f"{indented}Sequential object",
            f"{indented}Description: \"{self.get_description()}\"",
            f"{indented}Number of equations: {self.num_equations}",
            f"{indented}Number of [nonidentities, identities]: [{len(self.nonidentity_index)}, {len(self.identity_index)}]",
            f"{indented}Number of RHS-only names (excluding residuals): {len(self.rhs_only_names)}",
            f"{indented}[Min, Max] time shift: [{self.min_shift:+g}, {self.max_shift:+g}]",
        ))

    def precopy(self, /, ) -> None:
        """
        """
        for x in self.explanatories:
            x.precopy()

    #
    # ===== Implement SimulatableProtocol =====
    # This protocal is used in Dataslabs to prepare data arrays
    #

    def get_min_max_shifts(
        self,
        /,
    ) -> tuple[int, int]:
        """
        """
        return self.min_shift, self.max_shift

    def get_databank_names(
        self,
        plan: _plans.Plan | None = None,
        /,
    ) -> tuple[str]:
        """
        """
        all_names = self.all_names
        if plan is not None:
            plan_names = set(plan.collect_databank_names())
            extra_plan_names = tuple(plan_names.difference(all_names))
            all_names = all_names + extra_plan_names
        return all_names

    #
    # ===== Implement PlannableProtocol =====
    #

    def can_be_exogenized(self, /, ) -> tuple[str]:
        return tuple(
            n for i, n in enumerate(self.lhs_names)
            if not self.explanatories[i].is_identity
        )

    def can_be_exogenized_when_data(self, /, ) -> tuple[str]:
        return self.can_be_endogenized()

    def can_be_endogenized(self, /, ) -> None:
        return None
    #]





"""
"""


#[
from __future__ import annotations

from typing import (Literal, Callable, )
from collections.abc import (Iterable, )
from numbers import (Number, )
import functools as _ft
import json as _js
import numpy as _np

from .. import equations as _equations
from .. import quantities as _quantities
from .. import sources as _sources
from .. import dates as _dates
from ..incidences import main as _incidence
from ..databanks import main as _databanks
from ..fords import descriptors as _descriptors

from . import _flags
#]


"""
Quantities that are time series in model databanks
"""
_TIME_SERIES_QUANTITY = _quantities.QuantityKind.VARIABLE | _quantities.QuantityKind.SHOCK


_DATABANK_OUTPUT_FORMAT_RESOLUTION = {
    "dict": lambda x: x,
    "Databank": lambda x: _databanks.Databank.from_dict(x),
    "databank": lambda x: _databanks.Databank.from_dict(x),
    "json": lambda x: js_.dumps(x),
    "json4": lambda x: js_.dumps(x, indent=4),
}


def _decorate_output_format(func: Callable, ):
    """
    """
    #[
    def _wrapper(*args, **kwargs):
        output = func(*args, **kwargs)
        output_format = kwargs.get("output", "Databank")
        return _DATABANK_OUTPUT_FORMAT_RESOLUTION[output_format](output)
    return _wrapper
    #]


class GetMixin:
    """
    Frontend getter methods for Model objects
    """
    #[
    @_decorate_output_format
    def get_steady_levels(
        self,
        /,
        **kwargs,
    ) -> dict[str, Number]:
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _sources.LOGLY_VARIABLE, )
        return self._get_values("levels", qids, **kwargs, )

    @_decorate_output_format
    def get_steady_changes(
        self,
        /,
        **kwargs,
    ) -> dict[str, Number]:
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _sources.LOGLY_VARIABLE, )
        return self._get_values("changes", qids, **kwargs, )

    @_decorate_output_format
    def get_steady(
        self,
        /,
        **kwargs,
    ) -> dict[str, Number]:
        """
        """
        qids = tuple(_quantities.generate_qids_by_kind(self._invariant.quantities, _sources.LOGLY_VARIABLE, ))
        levels = self._get_values("levels", qids, **kwargs, )
        changes = self._get_values("changes", qids, **kwargs, )
        if self.is_singleton:
            out = {
                k: (levels[k], changes[k], )
                for k in levels.keys()
            }
        else:
            out = { k: None for k in levels.keys() }
            for k in levels.keys():
                out[k] = list(zip(levels[k], changes[k]))
        return out

    @_decorate_output_format
    def get_parameters(
        self,
        /,
        **kwargs,
    ) -> dict[str, Number]:
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.PARAMETER, )
        return self._get_values("levels", qids, **kwargs, )

    @_decorate_output_format
    def get_stds(
        self,
        /,
        **kwargs,
    ) -> dict[str, Number]:
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.STD, )
        return self._get_values("levels", qids, **kwargs, )

    @_decorate_output_format
    def get_parameters_stds(
        self,
        /,
        **kwargs,
    ) -> dict[str, Number]:
        qids = _quantities.generate_qids_by_kind(self._invariant.quantities, _quantities.QuantityKind.PARAMETER_OR_STD, )
        return self._get_values("levels", qids, **kwargs, )

    @_decorate_output_format
    def get_log_status(
        self,
        /,
    ) -> dict[str, bool]:
        return {
            qty.human: qty.logly
            for qty in self._invariant.quantities
            if qty.kind in _sources.LOGLY_VARIABLE
        }

    def get_initials(
        self,
        /,
    ) -> _incidence.Tokens:
        """
        Get required list of initial conditions
        """
        # Get the tokens of state vector with actual incidence in the
        # model, and lag them by one period to get initial conditions
        initial_tokens = [
            _incidence.Token(t.qid, t.shift-1)
            for t in self._invariant.dynamic_descriptor.solution_vectors.get_initials()
        ]
        return _incidence.print_tokens(initial_tokens, self.create_qid_to_name(), )

    def _get_steady_databank(
        self,
        start_date: _dates.Dater,
        end_date: _dates.Dater,
        /,
        deviation: bool = False,
    ) -> _databanks.Databank:
        """
        """
        num_columns = int(end_date - start_date + 1)
        shift_in_first_column = start_date.get_distance_from_origin()
        #
        array = self.create_some_array(
            deviation=deviation,
            num_columns=num_columns,
            shift_in_first_column=shift_in_first_column,
        )
        #
        qid_to_kind = self.create_qid_to_kind()
        remove_qids = tuple(
            qid
            for qid, kind in qid_to_kind.items()
            if kind not in _TIME_SERIES_QUANTITY
        )
        qid_to_name = {
            qid: name for qid, name in self.create_qid_to_name().items()
            if qid not in remove_qids
        }
        array = _np.delete(array, remove_qids, axis=0, )
        qid_to_description = self.create_qid_to_description()
        #
        return _databanks.Databank.from_array(
            array, qid_to_name, start_date,
            array_orientation="horizontal",
            interpret_dates="start_date",
            qid_to_description=qid_to_description,
        )

    def get_solution_vectors(self, /, ) -> _descriptors.SolutionVectors:
        """
        Get the solution vectors of the model
        """
        return self._invariant.dynamic_descriptor.solution_vectors

    def get_all_solution_matrices(self, /, ):
        return tuple(v.solution for v in self._variants)

    def get_solution_matrices(self, /, ):
        return self._variants[0].solution

    def get_dynamic_equations(
        self,
        /,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[_equations.Equation]:
        return tuple(
            _equations.generate_equations_of_kind(self._invariant.dynamic_equations, kind)
            if kind else self._invariant.dynamic_equations
        )

    def get_steady_equations(
        self,
        /,
        kind: _equations.EquationKind | None = None,
    ) -> tuple[_equations.Equation]:
        return tuple(
            _equations.generate_equations_of_kind(self._invariant.steady_equations, kind)
            if kind else self._invariant.steady_equations
        )

    def get_quantities(
        self,
        /,
        *,
        kind: _quantities.QuantityKind | None = None,
    ) -> _quantities.Quantities:
        return tuple(
            _quantities.generate_quantities_of_kind(self._invariant.quantities, kind)
            if kind else self._invariant.quantities
        )

    def get_flags(
        self,
        /,
    ) -> _flags.Flags:
        return self._invariant._flags

    def get_custom_functions(
        self,
        /,
    ) -> dict[str, Callable]:
        return self._invariant._function_context

    def _get_values(
        self,
        variant_attr: Literal["levels", "changes"],
        qids: Iterable[int],
        /,
        **kwargs,
    ) -> dict[str, Any]:
        """
        """
        custom_round = kwargs.pop("round", None, )
        def apply(value: Number) -> Number:
            value = float(value)
            return round(value, custom_round) if custom_round is not None else value
        #
        if self.is_singleton:
            def insert_value(out_dict, name, value):
                out_dict[name] = apply(value)
        else:
            def insert_value(out_dict, name, value):
                out_dict.setdefault(name, []).append(apply(value))
        #
        qids = tuple(qids)
        qid_to_name = self.create_qid_to_name()
        out_dict = {}
        #
        for v in self._variants:
            values = v.retrieve_values(variant_attr, qids)
            for qid, value in zip(qids, values):
                insert_value(out_dict, qid_to_name[qid], value)
        return out_dict
    #]




"""
Data arrays with row and column names
"""


#[
from __future__ import annotations

import numpy as _np
from typing import (Self, Protocol, )
from numbers import (Number, )
from collections.abc import (Iterable, )

from .series import main as _series
from .databanks import main as _databanks
from .plans import main as _plans
from . import dates as _dates
#]


__all__ = (
    "Dataslab",
)


class SimulatableProtocol(Protocol, ):
    def get_min_max_shift(self, /, ) -> tuple[int, int]:  ...
    def get_databank_names(self, /, ) -> tuple[str, ...]:  ...


class Dataslab:
    """
    """
    #[
    data: _np.ndarray | None = None
    row_names: Iterable[str] | None = None
    missing_names: tuple[str, ...] | None = None
    column_dates: tuple[Dater, ...] | None = None
    base_columns: tuple[int, ...] | None = None

    def to_databank(self, *args, **kwargs, ) -> _dates.Databank:
        """
        """
        return multiple_to_databank((self,), *args, **kwargs, )

    @classmethod
    def from_databank(
        cls,
        databank: _databanks.Databank,
        names: Iterable[str],
        ext_range: Ranger,
        /,
        column: int = 0,
    ) -> Self:
        """
        Add data from a databank to this dataslab
        """
        self = cls()
        missing_names = [
            n for n in names
            if n not in databank
        ]
        #
        num_periods = len(ext_range)
        nan_row = _np.full((1, num_periods), _np.nan, dtype=float)
        generate_data = tuple(
            _extract_data_from_record(databank[n], ext_range, column, )
            if n not in missing_names else nan_row
            for n in names
        )
        #
        self.data = _np.vstack(generate_data)
        self.row_names = tuple(names)
        self.column_dates = tuple(ext_range)
        self.missing_names = missing_names
        return self

    @classmethod
    def from_databank_for_simulation(
        cls,
        simulatable: SimulatableProtocol,
        databank: _databanks.Databank,
        base_range: Iterable[Dater],
        /,
        column: int = 0,
        plan: _plans.Plan | None = None,
    ) -> Self:
        """
        """
        ext_range, base_columns = (
            get_extended_range(
                simulatable,
                base_range,
            )
        )
        names = simulatable.get_databank_names(plan, )
        self = cls.from_databank(databank, names, ext_range, column=column, )
        self.base_columns = base_columns
        return self

    @property
    def num_ext_periods(self, /, ) -> int:
        """
        """
        return len(self.column_dates)

    @property
    def num_rows(self, /, ) -> int:
        """
        """
        return self.data.shape[0] if self.data is not None else 0

    def remove_terminal(self, /, ) -> None:
        """
        """
        last_base_column = self.base_columns[-1]
        self.data = self.data[:, :last_base_column+1]
        self.column_dates = self.column_dates[:last_base_column+1]

    def remove_columns(
        self,
        remove: int,
        /,
    ) -> None:
        """
        """
        if remove > 0:
            self.data = self.data[:, remove:]
            self.column_dates = self.column_dates[remove:]
        elif remove < 0:
            self.data = self.data[:, :remove]
            self.column_dates = self.column_dates[:remove]

    def copy_data(self, /, ) -> _np.ndarray:
        """
        """
        return self.data.copy()

    def fill_missing_in_base_columns(
        self,
        names,
        /,
        fill: Number = 0,
    ) -> None:
        """
        """
        base_slice = slice(self.base_columns[0], self.base_columns[-1]+1)
        for i, n in enumerate(self.row_names):
            if names is not Ellipsis and n not in names:
                continue
            values = self.data[i, base_slice]
            values[_np.isnan(values)] = fill

    def create_name_to_row(
        self,
        /,
    ) -> dict[str, int]:
        return { name: row for row, name in enumerate(self.row_names, ) }
    #]


def _extract_data_from_record(record, ext_range, column, /, ):
    """
    """
    return (
        record.get_data_column(ext_range, column).reshape(1, -1) 
        if hasattr(record, "get_data")
        else _np.full((1, len(ext_range)), float(record), dtype=float, )
    )


def multiple_to_databank(
    selves,
    /,
    target_databank: _databanks.Databank | None = None,
) -> _databanks.Databank:
    """
    Add data from a dataslab to a new or existing databank
    """
    #[
    if target_databank is None:
        target_databank = _databanks.Databank()
    self = selves[0]
    num_columns = len(selves)
    for row, n in enumerate(self.row_names):
        data = _np.hstack(tuple(ds.data[(row,), :].T for ds in selves))
        x = _series.Series(num_columns=num_columns, )
        x.set_data(self.column_dates, data)
        target_databank[n] = x
    return target_databank
    #]


def get_extended_range(
    simulatable: SimulatableProtocol,
    base_range: Iterable[_dates.Dater],
    /,
) -> Iterable[_dates.Dater]:
    """
    """
    base_range = tuple(t for t in base_range)
    num_base_periods = len(base_range)
    min_shift, max_shift = simulatable.get_min_max_shifts()
    if min_shift == 0:
        min_shift = -1
    start_date = base_range[0] + min_shift
    end_date = base_range[-1] + max_shift
    base_columns = tuple(range(-min_shift, -min_shift+num_base_periods))
    ext_range = tuple(_dates.Ranger(start_date, end_date))
    return ext_range, base_columns




"""
First-order system simulators
"""


#[
from __future__ import annotations

from typing import (Self, Any, TypeAlias, Literal, Protocol, runtime_checkable)
from collections.abc import (Iterable, )
import numpy as _np

from ..databanks import main as _databanks
from ..fords import simulators as _simulators
from .. import dataslabs as _dataslabs
#]


class SimulateMixin:
    """
    """
    #[
    def simulate(
        self,
        in_databank: _databanks.Databank,
        base_range: Iterable[Dater],
        /,
        anticipate: bool = True,
        deviation: bool = False,
        prepend_input: bool = True,
        add_to_databank: _databanks.Databank | None = None,
    ) -> tuple[_databanks.Databank, dict[str, Any]]:
        """
        """
        dataslabs = tuple(
            _dataslabs.Dataslab.from_databank_for_simulation(self, in_databank, base_range, column=i, )
            for i in range(self.num_variants)
        )
        #
        qid_to_logly = self.create_qid_to_logly()
        boolex_logly = tuple(
            qid_to_logly[qid] or False
            for qid in range(dataslabs[0].num_rows)
        )
        #
        for variant, dataslab in zip(self._variants, dataslabs):
            new_data = dataslab.copy_data()
            new_data = _simulators.simulate_flat(
                variant.solution, self.get_solution_vectors(), boolex_logly,
                new_data, dataslab.base_columns, deviation, anticipate,
            )
            dataslab.data = new_data
            dataslab.remove_columns(dataslab.base_columns[-1] - dataslab.num_ext_periods + 1)
        #
        out_db = _dataslabs.multiple_to_databank(dataslabs)
        if prepend_input:
            out_db.prepend(in_databank, base_range[0]-1, )
        out_db = out_db | self.get_parameters_stds()
        #
        # Add to custom databank
        #
        if add_to_databank is not None:
            out_db = add_to_databank | out_db
        #
        info = {"dataslabs": dataslabs, }
        return out_db, info
    #]


